## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl
## -- Module  : models_train.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-04-18  0.0.0     DA       Creation
## -- 2021-06-06  1.0.0     DA       Release of first version
## -- 2021-06-25  1.1.0     DA       Extension of classes Scenario and Training by data logging; 
## --                                New method Training.save_data()
## -- 2021-07-01  1.1.1     DA       Bugfixes in class Training
## -- 2021-07-06  1.1.2     SY       Update method Training.save_data()
## -- 2021-08-26  1.2.0     DA       New class HPTuningRL; incompatible changes on class Scenario
## -- 2021-08-28  1.2.1     DA       Bugfixes and minor improvements
## -- 2021-09-11  1.2.2     MRD      Change Header information to match our new library name
## -- 2021-10-05  1.2.3     SY       Bugfixes and minor improvements
## -- 2021-10-08  1.2.4     DA       Class Scenario/constructor/param p_cycle_limit: new value -1
## --                                lets class get the cycle limit from the env
## -- 2021-10-28  1.2.5     DA       Bugfix method Scenario.reset(): agent's buffer was not cleared
## -- 2021-11-13  1.3.0     DA       Rework/improvement of class RLTraining
## -- 2021-12-07  1.3.1     DA       - Method RLScenario.__init__(): param p_cycle_len removed
## --                                - Method RLTraining.__init__(): par p_scenario replaced by p_scenario_cls
## -- 2021-12-09  1.3.2     DA       Class RLTraining: introduced dynamic parameters **p_kwargs
## -- 2021-12-12  1.4.0     DA       Class RLTraining: evaluation and stagnation detection added
## -- 2021-12-16  1.4.1     DA       Method RLTraining._close_evaluation(): optimized scoring
## -- 2021-12-21  1.5.0     DA       Class RLTraining: reworked evaluation strategy
## -- 2022-01-19  1.5.1     DA       Corrections of log texts
## -- 2022-01-23  1.5.2     DA       Classes RLTrainingResults, RLDataStoringEval: added columns 
## --                                'Score (MA)' and 'Adaptations' to result file evaluation.csv
## -- 2022-01-26  1.5.3     DA       Class RLTraining: turn off training data logging during evaluation
## -- 2022-01-26  1.5.4     SY       Class RLTraining: update preparation of data logging for next episode
## -- 2022-01-28  1.5.5     DA       Class RLTraining: correction of stagnation detection
## -- 2022-02-04  1.6.0     DA       Class RLTraining: new parameter p_stagnation_entry
## -- 2022-02-10  1.7.0     DA       Class RLTraining: 
## --                                - new parameter p_end_at_stagnation
## --                                - file evaluation.csv: new columns 'Score until Stagnation', 
## --                                  'Score(MA) until Stagnation'
## -- 2022-02-27  1.7.1     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-05-23  1.7.2     SY       Bug fixing: storing data reward
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.7.2 (2022-05-23)

This module provides model classes to define and run rl scenarios and to train agents inside them.
"""

from mlpro.rl.models_env import *



## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLDataStoring(DataStoring):
    """
    Derivative of basic class DataStoring that is specialized to store episodic training data in the
    context of reinforcement learning.
    """

    # Frame ID renamed
    C_VAR0 = 'Episode ID'

    # Variables for episodic detail data storage
    C_VAR_CYCLE = 'Cycle'
    C_VAR_DAY = 'Day'
    C_VAR_SEC = 'Second'
    C_VAR_MICROSEC = 'Microsecond'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_space: Set = None):
        """
        Parameters:
            p_space         Space object that provides dimensional information for raw data. If None
                            a training header data object will be instantiated.
        """

        self.space = p_space

        # Initialization as an episodic detail data storage
        self.variables = [self.C_VAR_CYCLE, self.C_VAR_DAY, self.C_VAR_SEC, self.C_VAR_MICROSEC]
        self.var_space = []

        for dim_id in self.space.get_dim_ids():
            dim = self.space.get_dim(dim_id)
            self.var_space.append(dim.get_name_short())

        self.variables.extend(self.var_space)

        super().__init__(self.variables)

    ## -------------------------------------------------------------------------------------------------
    def get_variables(self):
        return self.variables

    ## -------------------------------------------------------------------------------------------------
    def get_space(self):
        return self.space

    ## -------------------------------------------------------------------------------------------------
    def add_episode(self, p_episode_id):
        self.add_frame(p_episode_id)
        self.current_episode = p_episode_id

    ## -------------------------------------------------------------------------------------------------
    def memorize_row(self, p_cycle_id, p_tstamp: timedelta, p_data):
        """
        Memorizes an episodic data row.

        Parameters: 
            p_cycle_id          Cycle id
            p_tstamp            Time stamp
            p_data              Data that meet the dimensionality of the related space
        """

        self.memorize(self.C_VAR_CYCLE, self.current_episode, p_cycle_id)
        self.memorize(self.C_VAR_DAY, self.current_episode, p_tstamp.days)
        self.memorize(self.C_VAR_SEC, self.current_episode, p_tstamp.seconds)
        self.memorize(self.C_VAR_MICROSEC, self.current_episode, p_tstamp.microseconds)

        for i, var in enumerate(self.var_space):
            self.memorize(var, self.current_episode, p_data[i])


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLDataStoringEval(DataStoring):
    """
    Derivative of basic class DataStoring that is specialized to store evaluation data of a training
    in the context of reinforcement learning.

    Parameters
    ----------
    p_space : Set
        Set object that provides dimensional information for raw data. If None a training header data object will
    be instantiated.

    """

    # Frame ID renamed
    C_VAR0                      = 'Evaluation ID'

    # Variables for training header data storage
    C_VAR_SCORE                 = 'Score'                       # Score
    C_VAR_SCORE_MA              = 'Score(MA)'                   # Moving average score
    C_VAR_SCORE_UNTIL_STAG      = 'Score until Stagnation'      # Score until first stagnation
    C_VAR_SCORE_MA_UNTIL_STAG   = 'Score(MA) until Stagnation'  # Moving average score unitl first stagnation
    C_VAR_NUM_CYCLES            = 'Cycles'                      # Number of cycles
    C_VAR_NUM_SUCCESS           = 'Successes'                   # Number of successful labeled states
    C_VAR_NUM_BROKEN            = 'Crashes'                     # Number of broken states
    C_VAR_NUM_LIMIT             = 'Timeouts'                    # Number of timeouts (cycles per episode reached)
    C_VAR_NUM_ADAPT             = 'Adaptations'                 # Number of adaptations in the last training period

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_space: Set):

        self.space = p_space

        # Initialization as an episodic detail data storage
        self.variables = [self.C_VAR_SCORE, 
                          self.C_VAR_SCORE_MA, 
                          self.C_VAR_SCORE_UNTIL_STAG, 
                          self.C_VAR_SCORE_MA_UNTIL_STAG, 
                          self.C_VAR_NUM_CYCLES, 
                          self.C_VAR_NUM_LIMIT,
                          self.C_VAR_NUM_SUCCESS, 
                          self.C_VAR_NUM_BROKEN, 
                          self.C_VAR_NUM_ADAPT]
        self.var_space = []

        for dim_id in self.space.get_dim_ids():
            dim = self.space.get_dim(dim_id)
            self.var_space.append(dim.get_name_short())

        self.variables.extend(self.var_space)

        super().__init__(self.variables)

    ## -------------------------------------------------------------------------------------------------
    def get_variables(self):
        return self.variables

    ## -------------------------------------------------------------------------------------------------
    def get_space(self):
        return self.space

    ## -------------------------------------------------------------------------------------------------
    def add_evaluation(self, p_evaluation_id):
        self.add_frame(p_evaluation_id)
        self.current_evaluation = p_evaluation_id

    ## -------------------------------------------------------------------------------------------------
    def memorize_row(self, 
                     p_score, 
                     p_score_ma, 
                     p_num_limit, 
                     p_num_cycles, 
                     p_num_success, 
                     p_num_broken,
                     p_num_adaptations, 
                     p_reward,
                     p_score_until_stag=None, 
                     p_score_ma_until_stag=None, 
                     ):
        """
        Memorizes an evaluation data row.

        Parameters
        ----------
        p_score : float
            Score value of current evaluation.
        p_score_ma : float
            Moving average score value.
        p_num_limit : int
            Number of episodes in timeout.
        p_num_cycles : int
            Number of evaluation cycles.
        p_num_success : int
            Number of states that were labeled as successfully.
        p_num_broken : int
            Number of states that were labeled as broken.
        p_num_adaptations : int
            Number of adaptations in the last training period.
        p_reward : list
            Episode Reward
        p_score_until_stag : float
            Optional score value of current evaluation until first stagnation. Default = None.
        p_score_ma_until_stag : float
            Optional moving average score value until first stagnation. Default = None. 
        
        """

        self.memorize(self.C_VAR_SCORE, self.current_evaluation, p_score)
        self.memorize(self.C_VAR_SCORE_MA, self.current_evaluation, p_score_ma)
        self.memorize(self.C_VAR_NUM_LIMIT, self.current_evaluation, p_num_limit)
        self.memorize(self.C_VAR_NUM_CYCLES, self.current_evaluation, p_num_cycles)
        self.memorize(self.C_VAR_NUM_SUCCESS, self.current_evaluation, p_num_success)
        self.memorize(self.C_VAR_NUM_BROKEN, self.current_evaluation, p_num_broken)
        self.memorize(self.C_VAR_NUM_ADAPT, self.current_evaluation, p_num_adaptations)

        if p_score_until_stag is not None: 
            self.memorize(self.C_VAR_SCORE_UNTIL_STAG, self.current_evaluation, p_score_until_stag)
        else:
            self.memorize(self.C_VAR_SCORE_UNTIL_STAG, self.current_evaluation, '')

        if p_score_ma_until_stag is not None: 
            self.memorize(self.C_VAR_SCORE_MA_UNTIL_STAG, self.current_evaluation, p_score_ma_until_stag)
        else:
            self.memorize(self.C_VAR_SCORE_MA_UNTIL_STAG, self.current_evaluation, '')

        for i, var in enumerate(self.var_space):
            self.memorize(var, self.current_evaluation, p_reward[i])


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLScenario(Scenario):
    """
    Template class for an RL scenario consisting of an environment and an agent. 
    """

    C_TYPE = 'RL-Scenario'
    C_NAME = '????'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self,
                 p_mode=Mode.C_MODE_SIM,  # Operation mode (see class Mode)
                 p_ada: bool = True,  # Boolean switch for adaptivity of internal model
                 p_cycle_limit=0,  # Maximum number of cycles (0=no limit, -1=get from env)
                 p_visualize=True,  # Boolean switch for env/agent visualisation
                 p_logging=Log.C_LOG_ALL):  # Log level (see constants of class Log)

        # 1 Setup entire scenario
        self._env = None
        super().__init__(p_mode=p_mode, p_ada=p_ada, p_cycle_limit=p_cycle_limit, p_visualize=p_visualize,
                         p_logging=p_logging)
        if self._env is None:
            raise ImplementationError('Please bind your RL environment to self._env')

        self._agent = self._model

        # 2 Finalize cycle limit
        if self._cycle_limit == -1:
            self._cycle_limit = self._env.get_cycle_limit()

        # 3 Init data logging
        self.connect_data_logger()

    ## -------------------------------------------------------------------------------------------------
    def switch_logging(self, p_logging):
        super().switch_logging(p_logging)
        self._env.switch_logging(p_logging)

    ## -------------------------------------------------------------------------------------------------
    def _setup(self, p_mode, p_ada: bool, p_logging: bool) -> Model:
        """
        Set up the ML scenario by redefinition. Please bind your environment to self._env and return
        the agent as model. 

        Parameters:
            p_mode          Operation mode (see class Mode)
            p_ada           Boolean switch for adaptivity of internal model
            p_logging       Boolean switch for logging functionality

        Returns:
            Agent model (object of type Agent or Multi-agent)
        """

        raise NotImplementedError

    ## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        super().init_plot(p_figure=p_figure)
        self._env.init_plot(p_figure=p_figure)

    ## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        super().update_plot()
        self._env.update_plot()

    ## -------------------------------------------------------------------------------------------------
    def _set_mode(self, p_mode):
        if isinstance(self._env, Mode):
            self._env.set_mode(p_mode)

    ## -------------------------------------------------------------------------------------------------
    def get_latency(self) -> timedelta:
        return self._env.get_latency()

    ## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed):
        """
        Environment and timer are reset. The random generators for environment and agent will
        also be reset. Optionally the agent's internal buffer data will be cleared, but
        it's policy will not be touched.

        Parameters:
            p_seed                  New seed for environment's and agent's random generator
        """

        # Reset environment
        self._env.reset(p_seed)
        if self._visualize:
            self._env.init_plot()

    ## -------------------------------------------------------------------------------------------------
    def get_agent(self):
        return self._agent

    ## -------------------------------------------------------------------------------------------------
    def get_env(self):
        return self._env

    ## -------------------------------------------------------------------------------------------------
    def connect_data_logger(self, p_ds_states: RLDataStoring = None, p_ds_actions: RLDataStoring = None,
                            p_ds_rewards: RLDataStoring = None):
        self._ds_states = p_ds_states
        self._ds_actions = p_ds_actions
        self._ds_rewards = p_ds_rewards

    ## -------------------------------------------------------------------------------------------------
    def _run_cycle(self):
        """
        Processes a single cycle.

        Returns
        -------
        success : bool
            True on success. False otherwise.
        error : bool
            True on error. False otherwise.
        adapted : bool
            True, if agent adapted something in this cycle. False otherwise.

        """

        # 1 Environment: get current state
        state = self._env.get_state()
        state.set_tstamp(self._timer.get_time())

        if self._ds_states is not None:
            self._ds_states.memorize_row(self._cycle_id, self._timer.get_time(), state.get_values())

        # 2 Agent: compute and log next action
        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Agent computes action...')
        action = self._agent.compute_action(state)
        ts = self._timer.get_time()
        action.set_tstamp(ts)
        if self._ds_actions is not None:
            self._ds_actions.memorize_row(self._cycle_id, ts, action.get_sorted_values())

        # 3 Environment: process agent's action
        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Env processes action...')
        self._env.process_action(action)
        self._timer.add_time(self._env.get_latency())  # in virtual mode only...
        self._env.get_state().set_tstamp(self._timer.get_time())

        # 4 Environment: compute and log reward
        reward = self._env.compute_reward()
        ts = self._timer.get_time()
        reward.set_tstamp(ts)
        if self._ds_rewards is not None:
            if (reward.get_type() == Reward.C_TYPE_OVERALL) or (reward.get_type() == Reward.C_TYPE_EVERY_AGENT):
                reward_values = np.zeros(self._ds_rewards.get_space().get_num_dim())

                for i, agent_id in enumerate(self._ds_rewards.get_space().get_dim_ids()):
                    reward_values[i] = reward.get_agent_reward(i)

                self._ds_rewards.memorize_row(self._cycle_id, ts, reward_values)

        # 5 Agent: adapt policy
        self.log(self.C_LOG_TYPE_I, 'Process time', self._timer.get_time(), ': Agent adapts policy...')
        adapted = self._agent.adapt(self._env.get_state(), reward)

        # 6 Check for terminating events
        success = self._env.get_state().get_success()
        error = self._env.get_state().get_terminal()

        if success:
            self.log(self.C_LOG_TYPE_S, 'Process time', self._timer.get_time(), ': Environment goal achieved')

        if error:
            self.log(self.C_LOG_TYPE_E, 'Process time', self._timer.get_time(), ': Environment terminated')

        return success, error, adapted


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLTrainingResults(TrainingResults):
    """
    Results of a RL training.

    Parameters
    ----------
    p_scenario : RLScenario
        Related reinforcement learning scenario.
    p_run : int
        Run id.
    p_cycle_id : int
        Id of first cycle of this run.
    p_path : str
        Optional destination path to store the results.
    p_logging
        Log level (see constants of class Log). Default: Log.C_LOG_ALL

    """

    C_NAME = 'RL'

    C_FNAME_EVAL = 'evaluation'
    C_FNAME_ENV_STATES = 'env_states'
    C_FNAME_AGENT_ACTIONS = 'agent_actions'
    C_FNAME_ENV_REWARDS = 'env_rewards'

    C_CPAR_NUM_EPI = 'Training Episodes'
    C_CPAR_NUM_EVAL = 'Evaluations'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_scenario: RLScenario, p_run, p_cycle_id, p_path=None, p_logging=Log.C_LOG_WE):
        super().__init__(p_scenario, p_run, p_cycle_id, p_path=p_path, p_logging=p_logging)

        self.num_episodes = 0
        self.num_evaluations = 0
        self.ds_states = None
        self.ds_actions = None
        self.ds_rewards = None
        self.ds_eval = None

    ## -------------------------------------------------------------------------------------------------
    def close(self):
        super().close()

        self.add_custom_result(self.C_CPAR_NUM_EPI, self.num_episodes)
        self.add_custom_result(self.C_CPAR_NUM_EVAL, self.num_evaluations)

    ## -------------------------------------------------------------------------------------------------
    def _log_results(self):
        super()._log_results()
        self.log(self.C_LOG_TYPE_W, '-- Training Episodes :', self.num_episodes)
        self.log(self.C_LOG_TYPE_W, '-- Evaluations       :', self.num_evaluations)

    ## -------------------------------------------------------------------------------------------------
    def save(self, p_path, p_filename='summary.csv') -> bool:
        if not super().save(p_path, p_filename=p_filename):
            return False

        if self.ds_states is not None:
            self.ds_states.save_data(p_path, self.C_FNAME_ENV_STATES)
        if self.ds_actions is not None:
            self.ds_actions.save_data(p_path, self.C_FNAME_AGENT_ACTIONS)
        if self.ds_rewards is not None:
            self.ds_rewards.save_data(p_path, self.C_FNAME_ENV_REWARDS)
        if self.ds_eval is not None:
            self.ds_eval.save_data(p_path, self.C_FNAME_EVAL)


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RLTraining(Training):
    """
    This class performs an episodic training on a (multi-)agent in a given environment. Both are
    expected as parts of a reinforcement learning scenario (see class RLScenario for more details).
    The class optionally collects all relevant data like environment states and rewards or agents
    actions. Furthermore, overarching training data will be collected.

    Parameters
    ----------
    p_scenario_cls 
        Name of RL scenario class, compatible to/inherited from class RLScenario.
    p_cycle_limit : int
        Maximum number of training cycles (0=no limit). Default = 0.
    p_cycles_per_epi_limit : int
        Optional limit of cycles per episode (0=no limit, -1=get environment limit). Default = -1.    
    p_adaptation_limit : int
        Maximum number of adaptations (0=no limit). Default = 0.
    p_eval_frequency : int
        Optional evaluation frequency (0=no evaluation). Default = 0.
    p_eval_grp_size : int
        Number of evaluation episodes (eval group). Default = 0.
    p_score_ma_horizon : int
        Horizon length for moving average score computation. Default = 5.
    p_stagnation_limit : int
        Optional limit of consecutive evaluations without training progress. Base is the moving average
        score. Default = 0.
    p_stagnation_entry : int
        Optional number of evaluations before the stagnation detection starts. Default = 0.
    p_end_at_stagnation : bool
        If True, the training ends when stagnation has beed detected. Default = True.
    p_hpt : HyperParamTuner
        Optional hyperparameter tuner (see class mlpro.bf.ml.HyperParamTuner). Default = None.
    p_hpt_trials : int
        Optional number of hyperparameter tuning trials. Default = 0. Must be > 0 if p_hpt is supplied.
    p_path : str
        Optional destination path to store training data. Default = None.
    p_collect_states : bool
        If True, the environment states will be collected. Default = True.
    p_collect_actions : bool
        If True, the agent actions will be collected. Default = True.
    p_collect_rewards : bool
        If True, the environment reward will be collected. Default = True.
    p_collect_eval : bool
        If True, global evaluation data will be collected. Default = True.
    p_visualize : bool
        Boolean switch for env/agent visualisation. Default = False.
    p_logging
        Log level (see constants of class mlpro.bf.various.Log). Default = Log.C_LOG_WE.

    """

    C_NAME = 'RL'

    C_CLS_RESULTS = RLTrainingResults

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, **p_kwargs):

        # 1 Initialization of elementary training functionalities
        super().__init__(**p_kwargs)

        # 2 Check and completion of RL-specific parameters

        # 2.1 Optional parameter p_cycles_per_epi_limit
        try:
            self._cycles_per_epi_limit = self._kwargs['p_cycles_per_epi_limit']
        except KeyError:
            self._cycles_per_epi_limit = -1

        # 2.2 Optional parameter p_eval_frequency
        try:
            self._eval_frequency = self._kwargs['p_eval_frequency']
        except KeyError:
            self._eval_frequency = 0
            self._kwargs['p_eval_frequency'] = self._eval_frequency

        # 2.3 Optional parameter p_eval_grp_size
        try:
            self._eval_grp_size = self._kwargs['p_eval_grp_size']
        except KeyError:
            self._eval_grp_size = 0
            self._kwargs['p_eval_grp_size'] = self._eval_grp_size

        # 2.4 Optional parameter p_ma_score_horizon
        try:
            self._eval_score_ma_horizon = self._kwargs['p_score_ma_horizon']
            if self._eval_score_ma_horizon < 1:
                self._eval_score_ma_horizon = 1

        except KeyError:
            self._eval_score_ma_horizon = 5
            self._kwargs['p_score_ma_horizon'] = self._eval_score_ma_horizon

        # 2.5 Optional parameter p_stagnation_limit
        try:
            self._stagnation_limit = self._kwargs['p_stagnation_limit']
        except KeyError:
            self._stagnation_limit = 0
            self._kwargs['p_stagnation_limit'] = self._stagnation_limit

        # 2.6 Optional parameter p_stagnation_entry
        try:
            self._stagnation_entry = self._kwargs['p_stagnation_entry']
        except KeyError:
            self._stagnation_entry = 0
            self._kwargs['p_stagnation_entry'] = self._stagnation_entry

        # 2.7 Optional parameter p_end_at_stagnation
        try:
            self._end_at_stagnation = self._kwargs['p_end_at_stagnation']
        except KeyError:
            self._end_at_stagnation = True
            self._kwargs['p_end_at_stagnation'] = self._end_at_stagnation

        # 2.8 Optional parameter p_collect_states
        try:
            self._collect_states = self._kwargs['p_collect_states']
        except KeyError:
            self._collect_states = True
            self._kwargs['p_collect_states'] = self._collect_states

        # 2.9 Optional parameter p_collect_actions
        try:
            self._collect_actions = self._kwargs['p_collect_actions']
        except KeyError:
            self._collect_actions = True
            self._kwargs['p_collect_actions'] = self._collect_actions

        # 2.10 Optional parameter p_collect_rewards
        try:
            self._collect_rewards = self._kwargs['p_collect_rewards']
        except KeyError:
            self._collect_rewards = True
            self._kwargs['p_collect_rewards'] = self._collect_rewards

        # 2.11 Optional parameter p_collect_eval
        try:
            self._collect_eval = self._kwargs['p_collect_eval']
        except KeyError:
            self._collect_eval = True
            self._kwargs['p_collect_eval'] = self._collect_eval

        # 3 Check for further restrictions
        if (self._cycle_limit <= 0) and (self._adaptation_limit <= 0) and (self._stagnation_limit <= 0):
            raise ParamError(
                'Please define at least one termination criterion (p_cycle_limit, p_adaptation_limit, p_stagnation_limit')

        if (self._stagnation_limit > 0) and ((self._eval_frequency <= 0) or (self._eval_grp_size <= 0)):
            raise ParamError(
                'For stagnation detection both parameters p_eval_frequency and p_eval_grp_size must be > 0 as well')

        if ((self._eval_frequency > 0) and (self._eval_grp_size <= 0)) or (
                (self._eval_frequency <= 0) and (self._eval_grp_size > 0)):
            raise ParamError('For cyclic evaluation both parameters p_eval_frequency and p_eval_grp_size must be > 0')

        if ( self._stagnation_entry > 0 ) and ( self._stagnation_limit <= 0 ):
            raise ParamError('Parameter p_stagnation_entry > 0 has no effect, because stagnation detection is off')

            # 4 Initialization of further rl-specific attributes
        if self._scenario is not None:
            self._mode = self.C_MODE_TRAIN
            self._cycles_episode = 0
            self._eval_grp_id = 0
            self._eval_last_score = None
            self._eval_stagnations = 0
            self._eval_stagnation_detected  = False
            self._eval_num_adaptations = 0
            self._eval_score_ma_buffer = []
            self._eval_score_ma_sum = 0

            self._env = self._scenario.get_env()
            self._agent = self._scenario.get_agent()

            if self._cycles_per_epi_limit == -1:
                self._cycles_per_epi_limit = self._env.get_cycle_limit()

            if self._cycles_per_epi_limit <= 0:
                raise ParamError(
                    'Please define a maximum number of training cycles per episode (env or param p_cycles_per_epi_limit')

            if self._eval_frequency > 0:
                # Training with evaluation starts with initial evaluation
                self._counter_epi_train = 0
                self._counter_epi_eval = 0
                self._mode = self.C_MODE_EVAL

    ## -------------------------------------------------------------------------------------------------
    def _init_results(self) -> TrainingResults:
        results = super()._init_results()

        if self._collect_states:
            results.ds_states = RLDataStoring(self._env.get_state_space())

        if self._collect_actions:
            results.ds_actions = RLDataStoring(self._env.get_action_space())

        if self._collect_rewards or self._collect_eval:
            reward_type = self._env.get_reward_type()

            if (reward_type == Reward.C_TYPE_OVERALL) or (reward_type == Reward.C_TYPE_EVERY_AGENT):
                reward_space = Set()
                try:
                    agents = self._agent.get_agents()
                except:
                    agents = [[self._agent, 1.0]]

                for agent, weight in agents:
                    reward_space.add_dim(Dimension(agent.get_name()))

                if self._collect_rewards:
                    results.ds_rewards = RLDataStoring(reward_space)

                if self._collect_eval:
                    results.ds_eval = RLDataStoringEval(reward_space)

        self._scenario.connect_data_logger(p_ds_states=results.ds_states, p_ds_actions=results.ds_actions,
                                           p_ds_rewards=results.ds_rewards)

        return results

    ## -------------------------------------------------------------------------------------------------
    def _init_episode(self):

        # 1 Evaluation handling  
        if self._eval_frequency > 0:

            if self._mode == self.C_MODE_TRAIN:

                if self._counter_epi_train == 0:
                    self._scenario.get_model().switch_adaptivity(True)

                    self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
                    self.log(self.C_LOG_TYPE_W, '-- Training period started...')
                    self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n')

                self._scenario.reset(self._results.num_episodes + self._eval_grp_size)

                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
                self.log(self.C_LOG_TYPE_W, '-- Training episode', self._results.num_episodes, 'started...')
                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n')

            else:
                if self._counter_epi_eval == 0:
                    self._scenario.get_model().switch_adaptivity(False)

                    self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
                    self.log(self.C_LOG_TYPE_W, '-- Evaluation period', self._results.num_evaluations, 'started...')
                    self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n')

                    self._init_evaluation()

                self._scenario.reset(self._counter_epi_eval)

                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
                self.log(self.C_LOG_TYPE_W, '-- Evaluation episode', self._counter_epi_eval, 'started...')
                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n')

        else:
            self._scenario.reset(self._results.num_episodes + self._eval_grp_size)

            self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, '-- Training episode', self._results.num_episodes, 'started...')
            self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n')

        # 3 Preparation of data logging for next episode
        if (self._results.ds_states and self._scenario._ds_states) is not None:
            self._results.ds_states.add_episode(self._results.num_episodes)
        if (self._results.ds_actions and self._scenario._ds_actions) is not None:
            self._results.ds_actions.add_episode(self._results.num_episodes)
        if (self._results.ds_rewards and self._scenario._ds_rewards) is not None:
            self._results.ds_rewards.add_episode(self._results.num_episodes)

    ## -------------------------------------------------------------------------------------------------
    def _close_episode(self):

        if self._eval_frequency > 0:

            if self._mode == self.C_MODE_TRAIN:
                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
                self.log(self.C_LOG_TYPE_W, '-- Training episode', self._results.num_episodes, 'finished after',
                         str(self._cycles_episode), 'cycles')
                self.log(self.C_LOG_TYPE_W, '-- Training cycles finished:', self._results.num_cycles + 1)
                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n\n')

                self._results.num_episodes += 1
                self._counter_epi_train += 1

                if self._counter_epi_train >= self._eval_frequency:
                    self._counter_epi_eval = 0
                    self._mode = self.C_MODE_EVAL

            else:
                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
                self.log(self.C_LOG_TYPE_W, '-- Evaluation episode', self._counter_epi_eval, 'finished after',
                         str(self._cycles_episode), 'cycles')
                self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n\n')

                self._counter_epi_eval += 1

                if self._counter_epi_eval >= self._eval_grp_size:

                    score = self._close_evaluation()
                    if (self._results.highscore is None) or (score > self._results.highscore):
                        self.log(self.C_LOG_TYPE_W, 'New temporal highscore', str(score))
                        self._results.highscore = score
                    else:
                        self.log(self.C_LOG_TYPE_W, 'New score', str(score))

                    self._results.num_evaluations += 1
                    self._counter_epi_train = 0
                    self._mode = self.C_MODE_TRAIN

        else:
            self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR)
            self.log(self.C_LOG_TYPE_W, '-- Training episode', self._results.num_episodes, 'finished after',
                     str(self._cycles_episode), 'cycles')
            self.log(self.C_LOG_TYPE_W, '-- Training cycles finished:', self._results.num_cycles + 1)
            self.log(self.C_LOG_TYPE_W, Training.C_LOG_SEPARATOR, '\n\n')

            self._results.num_episodes += 1

        self._cycles_episode = 0

    ## -------------------------------------------------------------------------------------------------
    def _init_evaluation(self):
        """
        Initializes the next evaluation.
        """

        # Training data logging is turned eff
        self._scenario.connect_data_logger()

        if self._results.ds_eval is not None:
            self._results.ds_eval.add_evaluation(self._results.num_evaluations)

        self._eval_num_cycles = 0
        self._eval_num_limit = 0
        self._eval_num_success = 0
        self._eval_num_broken = 0
        self._eval_sum_reward = None

    ## -------------------------------------------------------------------------------------------------
    def _update_evaluation(self, p_success: bool, p_error: bool, p_cycle_limit: bool):
        """
        Updates evaluation statistics.

        Parameters
        ----------
        p_success : bool
            True on success. False otherwise.
        p_error : bool
            True on error. False otherwise.
        p_cycle_limit : bool
            True, if cycle limit has reached. False otherwise.

        """

        self._eval_num_cycles += 1
        if p_cycle_limit:
            self._eval_num_limit += 1
        if p_success:
            self._eval_num_success += 1
        if p_error:
            self._eval_num_broken += 1

        reward = self._env.get_last_reward()
        if reward is None:
            return

        reward_type = reward.get_type()

        if reward_type == Reward.C_TYPE_OVERALL:
            if self._eval_sum_reward is None:
                self._eval_sum_reward = np.zeros(1)

            self._eval_sum_reward[0] += reward.get_overall_reward()

        elif reward_type == Reward.C_TYPE_EVERY_AGENT:
            if self._eval_sum_reward is None:
                self._eval_sum_reward = np.zeros(len(reward.agent_ids))

            for i, agent_id in enumerate(reward.agent_ids):
                # Get weight of agent
                try:
                    weight_agent = self._agent.get_agent(agent_id)[1]
                except:
                    weight_agent = 1.0

                self._eval_sum_reward[i] += reward.get_agent_reward(agent_id) * weight_agent

        else:
            raise Error('Reward type ' + str(reward_type) + ' not yet supported')

    ## -------------------------------------------------------------------------------------------------
    def _close_evaluation(self) -> np.ndarray:
        """
        Closes the current evaluation and computes a related score.

        Returns
        -------
        score : float
            Score of current evaluation.

        """

        # 1 Computation of score
        self._eval_sum_reward /= self._counter_epi_eval
        score = np.mean(self._eval_sum_reward)

        # 2 Computation of moving average score
        self._eval_score_ma_buffer.insert(0, score)
        self._eval_score_ma_sum += score

        if len(self._eval_score_ma_buffer) > self._eval_score_ma_horizon:
            self._eval_score_ma_sum -= self._eval_score_ma_buffer.pop()

        score_ma = self._eval_score_ma_sum / len(self._eval_score_ma_buffer)

        # 3 Store evaluation statistics
        if self._results.ds_eval is not None:
            num_adaptations = self._results.num_adaptations - self._eval_num_adaptations
            self._eval_num_adaptations = self._results.num_adaptations

            if not self._eval_stagnation_detected:
                score_until_stag    = score
                score_ma_until_stag = score_ma
            else:
                score_until_stag    = None
                score_ma_until_stag = None

            self._results.ds_eval.memorize_row(p_score=score,
                                               p_score_ma=score_ma,
                                               p_num_limit=self._eval_num_limit,
                                               p_num_cycles=self._eval_num_cycles,
                                               p_num_success=self._eval_num_success,
                                               p_num_broken=self._eval_num_broken,
                                               p_num_adaptations=num_adaptations,
                                               p_reward=self._eval_sum_reward,
                                               p_score_until_stag=score_until_stag,
                                               p_score_ma_until_stag=score_ma_until_stag)
                                            

        # 4 Stagnation detection based on moving average score
        if ( self._results.num_evaluations + 1 ) > self._stagnation_entry:

            if ( self._eval_last_score is not None ) and ( score_ma <= self._eval_last_score ):
                self._eval_stagnations += 1

            else:
                self._eval_stagnations = 0

            self._eval_last_score  = score_ma

        # 5 Training data logging is turned on
        self._scenario.connect_data_logger(p_ds_states=self._results.ds_states, p_ds_actions=self._results.ds_actions,
                                           p_ds_rewards=self._results.ds_rewards)

        # 6 Outro
        return score

    ## -------------------------------------------------------------------------------------------------
    def _run_cycle(self) -> bool:
        """
        Runs single training cycle.

        Returns:
            True, if training has finished. False otherwise.
        """

        # 0 Intro
        eof_episode = False
        eof_training = False

        # 1 Init next episode
        if self._cycles_episode == 0:
            self._init_episode()

        # 2 Run a cycle
        success, error, timeout, limit, adapted = self._scenario.run_cycle()
        self._cycles_episode += 1

        if adapted:
            self._results.num_adaptations += 1

        # 3 Update current evaluation
        if self._mode == self.C_MODE_EVAL:
            self._update_evaluation(p_success=success, p_error=error, p_cycle_limit=limit)

        # 4 Check: Episode finished?
        state = self._env.get_state()
        eof_episode = state.get_terminal()

        if eof_episode:

            if state.get_broken():
                # 4.1 Environment reached a state of no return
                self.log(self.C_LOG_TYPE_E, 'Environment broken')

            elif state.get_success():
                # 4.2 Objective of environment reached
                self.log(self.C_LOG_TYPE_W, 'Objective of environment reached')

            elif state.get_timeout():
                # 4.3 Cycle limit of environment reached
                self.log(self.C_LOG_TYPE_W, 'Limit of', self._env.get_cycle_limit(),
                         'cycles per episode reached (Environment)')

            else:
                # 4.4 Environment terminated the episode because of unknown reasons
                self.log(self.C_LOG_TYPE_W, 'Environment terminated episode (reason unknown)')

        elif self._cycles_episode == self._cycles_per_epi_limit:
            # 4.5 Cycle limit of training setup reached
            self.log(self.C_LOG_TYPE_W, 'Limit of', self._cycles_per_epi_limit, 'cycles per episode reached (Training)')
            eof_episode = True

        if eof_episode:
            self._close_episode()

        # 5 Check: Training finished?
        if (self._adaptation_limit > 0) and (self._results.num_adaptations == self._adaptation_limit):
            self.log(self.C_LOG_TYPE_W, 'Adaptation limit ', str(self._adaptation_limit), ' reached')
            eof_training = True

        if (self._stagnation_limit > 0) and ( self._eval_stagnations == self._stagnation_limit ):
            self.log(self.C_LOG_TYPE_W, 'Stagnation limit ', str(self._stagnation_limit), ' reached')
            self._eval_stagnations += 1              # Just to avoid further log lines for the same stagnation event
            self._eval_stagnation_detected = True
            eof_training = self._end_at_stagnation

        # 6 Outro
        return eof_training


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.bf.ui.scui
## -- Module  : main
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-04-13  0.0.0     DA       Creation
## -- 2021-06-20  0.5.0     DA       Pre-release with basic functionality
## -- 2021-07-05  0.6.0     DA       Pre-release with basic functionality
## -- 2021-09-11  0.6.0     MRD      Change Header information to match our new library name
## -------------------------------------------------------------------------------------------------

"""
Ver. 0.6.0 (2021-07-05)

Provides the SciUI application class.
"""


from tkinter import *
import tkinter.messagebox
from mlpro.bf.ui.sciui.framework import *





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class SciUI(SciUIWindow):

    C_TYPE          = 'Application'
    C_NAME          = 'SciUI' 
    C_VERSION       = '0.6.0'

## -------------------------------------------------------------------------------------------------
    def __init__(self, p_fullscreen=False, p_autoscan_scenarios=True, p_start_immediately=True, p_logging=False):
        """
        Parameters:
            p_fullscreen                If True, application starts in fullscreen mode.
            p_autoscan_screnarios       If True, all implemented scenario classes will be detected and 
                                        included.
            p_start_imediately          If True, application starts directly.
            p_logging                   Boolean switch for logging
        """

        super().__init__(p_logging=p_logging)

        # Init scenario handling
        self.scenario_pool      = []
        self.scenario_id        = IntVar()
        self.scenario_id.set(0)
        self.scenario           = None

        # Init shared db
        self.__init_shared_db()
        self.shared_db.window_fullscreen    = not p_fullscreen
        
        self.ignore_resize                  = True
        
        # Autoscan scenario classes
        if p_autoscan_scenarios:
            for cls in SciUIScenario.__subclasses__():
                self.register_scenario(p_scenario=cls, p_recursive=True, p_sort=False)  
                
            self.scenario_pool.sort()  
            
        # Start immediately, if requestsd
        if p_start_immediately: self.start()

        
## -------------------------------------------------------------------------------------------------
    def __init_shared_db(self): 
        self.shared_db = SciUISharedDB(self)
        self.shared_db.refresh_all.trace('w', self.__cb_global_refresh)

              
## -------------------------------------------------------------------------------------------------
    def __init_main_menu(self):       
        # 0 Init main menu
        menu = Menu(self, tearoff=False)
        self.config(menu=menu)
        
        # 1 Init Tk-Variables for menu control
        self.window_fullscreen = BooleanVar()
        self.window_fullscreen.set(self.shared_db.window_fullscreen)
        self.window_fullscreen.trace('w', self.__cb_toggle_fullscreen)
        
        # 2 Submenu 'File'
        file_menu = Menu(menu, tearoff=False)
        menu.add_cascade(label='File', menu=file_menu)
        # file_menu.add_command(label='Open Configuration', state='disabled', command=self.__cb_menu_file_open_configuration)
        # file_menu.add_command(label='Save Configuration', state='disabled', command=self.__cb_menu_file_save_configuration)
        # file_menu.add_separator()
        
        # 2.1 Submenu 'File/Change Scenario'
        scenario_menu = Menu(menu, tearoff=True)
        file_menu.add_cascade(label='Change Scenario', menu=scenario_menu)
        
        i = 0
        for scenario in self.scenario_pool:
            if scenario[1].C_RELEASED:
                state = 'active'
            else:
                state = 'disabled'
            scenario_menu.add_radiobutton(label=scenario[1].C_NAME + ' (Ver. ' + scenario[1].C_VERSION + ')', state=state, variable=self.scenario_id, value=i, command=self.__cb_submenu_file_change_scenario)
            i += 1
            
        file_menu.add_separator()
        
        # 2.2 Submenu 'File/Properties'
        file_menu.add_command(label='Properties', state='disabled', command=self.__cb_menu_file_properties)       
        file_menu.add_separator()
        
        file_menu.add_command(label='Exit', command=self.quit)

        # 3 Submenu 'View'
        view_menu = Menu(menu, tearoff=False)
        menu.add_cascade(label='View', menu=view_menu)
        view_menu.add_checkbutton(label='Fullscreen', onvalue=True, offvalue=False, variable=self.window_fullscreen)
        # view_menu.add_separator()
        # view_menu.add_command(label='Color Profile', state='disabled', command=self.__cb_menu_view_color_profile)
        
        # 4 Submenu 'Help'
        help_menu = Menu(menu, tearoff=False)
        menu.add_cascade(label='Help', menu=help_menu)
        help_menu.add_command(label='About...', command=self.__cb_menu_help_about)


## -------------------------------------------------------------------------------------------------
    def __init_main_window(self):
        self.ignore_resize = True
        self.title(self.C_NAME)

        try:
            self.attributes('-zoomed', True)
        except:
            self.state('zoomed')

        self.__set_fullscreen(not self.shared_db.window_fullscreen)  
        self.__init_main_menu()
        self.bind("<Configure>", self.__cb_window_resized)      
        

## -------------------------------------------------------------------------------------------------
    def start(self):
        
        # Check: at least one scenario registered?
        if len(self.scenario_pool) == 0: 
            self.log(self.C_LOG_TYPE_I, 'Please register at least one scenario')
            # exception...
            return 
        
        # Init main window
        self.__init_main_window()
        
        # Start application
        super().start()   
        

## -------------------------------------------------------------------------------------------------
    def register_scenario(self, p_scenario:SciUIScenario, p_recursive=False, p_sort=True): 
        self.__register_scenario_rec(p_scenario=p_scenario, p_recursive=p_recursive)
        if p_sort: self.scenario_pool.sort()      


## -------------------------------------------------------------------------------------------------
    def __register_scenario_rec(self, p_scenario:SciUIScenario, p_recursive): 
        if p_scenario.C_VISIBLE: 
            self.scenario_pool.append([p_scenario.C_NAME, p_scenario])
            self.log(self.C_LOG_TYPE_I, 'Scenario "' + p_scenario.C_NAME + ' Ver. (' + p_scenario.C_VERSION + ')" registered')

        if not p_recursive: return 
        
        for cls_child in p_scenario.__subclasses__():
            self.__register_scenario_rec(p_scenario=cls_child, p_recursive=p_recursive)


## -------------------------------------------------------------------------------------------------
    def __change_scenario(self):
        self.log(self.C_LOG_TYPE_W, '!!!')
        self.log(self.C_LOG_TYPE_W, 'GAP DEL SELF.SCENARIO')
        self.log(self.C_LOG_TYPE_W, '!!!')
#         del self.scenario
#         self.scenario = None
#         
        if self.scenario != None:
            # self.scenario.__del__()
            self.scenario = None  
              
        self.scenario         = self.scenario_pool[self.scenario_id.get()][1](self.shared_db, p_logging=self._level)
        self.shared_db.start_global_refresh()
        
        self.title(self.C_NAME + ': Scenario "' + self.scenario.C_NAME + '"')


## -------------------------------------------------------------------------------------------------
    def __set_fullscreen(self, p_fullscreen): 
        if self.shared_db.window_fullscreen == p_fullscreen: return
        self.log(self.C_LOG_TYPE_I, 'Main window fullscreen: ', p_fullscreen)
        self.attributes('-fullscreen', p_fullscreen)
        
        self.shared_db.window_fullscreen    = p_fullscreen
        self.shared_db.window_resized       = True
        

## -------------------------------------------------------------------------------------------------
    def __cb_global_refresh(self, *args):
        self.log(self.C_LOG_TYPE_I, 'Global Refresh')
        
        self.shared_db.refresh_full = True
        self.shared_db.redraw       = True
        if self.scenario != None: self.scenario.refresh()
        self.shared_db.refresh_full = False  #after_refresh()        


## -------------------------------------------------------------------------------------------------
    def __cb_toggle_fullscreen(self, *args):
        self.__set_fullscreen(self.window_fullscreen.get())        
        

## -------------------------------------------------------------------------------------------------
    def __cb_menu_file_properties(self):
        self.log(self.C_LOG_TYPE_I, 'Main Menu action: File/Properties')
        
        # Switch for Logging/Log Level


## -------------------------------------------------------------------------------------------------
    def __cb_menu_help_about(self):
        self.log(self.C_LOG_TYPE_I, 'Main Menu action: Help/About')    
        tkinter.messagebox.showinfo('About SciUI', 'SciUI - Scientific User Interface\nVersion ' + self.C_VERSION)


## -------------------------------------------------------------------------------------------------
    def __cb_submenu_file_change_scenario(self):
        self.log(self.C_LOG_TYPE_I, 'Sub Menu File/Change scenario:', self.scenario_id.get())
        self.__change_scenario()    


## -------------------------------------------------------------------------------------------------
    def __cb_window_resized(self, event):
        if event.widget != self: return
        if ( event.width == self.shared_db.window_width ) and ( event.height == self.shared_db.window_height ): return
        
        if self.ignore_resize:
            self.ignore_resize = False
            return

        self.log(self.C_LOG_TYPE_I, 'Main Window resized to {}x{}'.format(event.width, event.height))

        # Update screen and window properties
        self.shared_db.window_width     = event.width
        self.shared_db.window_height    = event.height

        # Rebuild active scenario
        self.shared_db.window_resized   = True
        self.__change_scenario()
        self.shared_db.window_resized   = False



## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro
## -- Module  : test_buffer
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-27  1.0.0     WB       Creation
## -- 2021-09-27  1.0.0     WB       Release First Version
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.0 (2021-09-27)

Unit test classes for SARBuffer.
"""


import pytest
import random
import numpy as np
from mlpro.bf.various import *
from mlpro.bf.math import *
from mlpro.bf.ml import *
from mlpro.rl.models import *
from mlpro.wrappers.openai_gym import WrEnvGYM2MLPro
from mlpro.rl.pool.sarsbuffer.PrioritizedBuffer import PrioritizedBuffer
from mlpro.rl.pool.sarsbuffer.RandomSARSBuffer import RandomSARSBuffer
from mlpro.rl.pool.policies.dummy import MyDummyPolicy
import gym
import random
from pathlib import Path



            
# -------------------------------------------------------------------------------------------------
@pytest.mark.parametrize("buffer_cls", [PrioritizedBuffer, RandomSARSBuffer])
def test_buffer(buffer_cls):    
    class MyScenario(RLScenario):

        C_NAME      = 'Matrix'

        def _setup(self, p_mode, p_ada, p_logging):
            gym_env     = gym.make('CartPole-v1')
            self._env   = WrEnvGYM2MLPro(gym_env, p_logging=False)

            class MyDummyPol(MyDummyPolicy):
                C_BUFFER_CLS = buffer_cls
                
            # 2 Setup standard single-agent with own policy
            return Agent(
                p_policy=MyDummyPol(
                    p_observation_space=self._env.get_state_space(),
                    p_action_space=self._env.get_action_space(),
                    p_buffer_size=100,
                    p_ada=p_ada,
                    p_logging=p_logging
                ),    
                p_envmodel=None,
                p_name='Smith',
                p_ada=p_ada,
                p_logging=p_logging
            )
            
    training        = RLTraining(
        p_scenario_cls=MyScenario,
        p_cycle_limit=100,
        p_max_stagnations=0,
        p_collect_states=True,
        p_collect_actions=True,
        p_collect_rewards=True,
        p_collect_training=True,
        p_logging=False
    )

    training.run()
    




## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.wrappers
## -- Module  : openai_gym.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-08-27  0.0.0     SY       Creation
## -- 2021-08-27  1.0.0     SY       Release of first version
## -- 2021-09-11  1.0.0     MRD      Change Header information to match our new library name
## -- 2021-09-23  1.1.0     SY       Update WrEnvGym to solve big data issues
## -- 2021-09-24  1.1.1     MRD      Change the gym wrapper _recognize_space() function to seperate
## --                                between discrete space and continuous space
## -- 2021-09-28  1.1.2     SY       WrEnvGym: implementation of method get_cycle_limits()
## -- 2021-09-29  1.1.3     SY       Change name: WrEnvGym to WrEnvGYM2MLPro
## -- 2021-09-30  1.2.0     SY       New classes: WrEnvMLPro2GYM
## -- 2021-10-05  1.2.1     SY       Update following new attributes done and broken in State
## -- 2021-10-06  1.2.2     DA       Minor fixes
## -- 2021-10-07  1.2.3     MRD      Redefine WrEnvMLPro2GYM reset(), step(), _recognize_space() function
## --                                Redefine also _recognize_space() from WrEnvGYM2MLPro
## -- 2021-10-08  1.2.4     DA       Correction of wrapper WREnvGYM2MLPro
## -- 2021-10-27  1.2.5     MRD      Remove reset() on WREnvGYM2MLPro and WrEnvMLPro2GYM init() function
## --                                to prevent double reset, due to it will be reset later on Training Class
## --                                Mismatch datatype last_done on WrPolicySB32MLPro
## -- 2021-11-13  1.2.6     DA       Minor adjustments
## -- 2021-11-16  1.2.7     DA       Refactoring
## -- 2021-11-16  1.2.8     SY       Refactoring
## -- 2021-12-03  1.2.9     DA       Refactoring
## -- 2021-12-21  1.3.0     DA       - Replaced 'done' by 'success' on mlpro functionality
## --                                - Optimized 'done' detection in both classed
## -- 2021-12-23  1.3.1     MRD      Remove adding self._num_cycle on simulate_reaction() due to 
## --                                EnvBase.process_actions() is already adding self._num_cycle
## -- 2022-01-21  1.3.2     DA/MRD   Class WrEnvMLPro2GYM: 
## --                                - refactored done detection 
## --                                - removed artifacts of cycle counting
## -- 2022-01-28  1.3.3     DA       Class WrEnvMLPro2GYM: stabilized destructor
## -- 2022-02-27  1.3.4     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-03-21  1.3.5     MRD      Added new parameter to the WrEnvMLPro2GYM.reset()
## -- 2022-05-19  1.3.6     SY       Gym 0.23: Replace function env.seed(seed) to env.reset(seed=seed)
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.3.6 (2022-05-19)

This module provides wrapper classes for reinforcement learning tasks.
"""

import gym
from mlpro.rl.models import *


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class WrEnvGYM2MLPro(Environment):
    """
    This class is a ready to use wrapper class for OpenAI Gym environments. 
    Objects of this type can be treated as an environment object. Encapsulated 
    gym environment must be compatible to class gym.Env.
    """

    C_TYPE = 'OpenAI Gym Env'

    ## -------------------------------------------------------------------------------------------------
    def __init__(self,
                 p_gym_env,  # Gym environment object
                 p_state_space: MSpace = None,  # Optional external state space object that meets the
                 # state space of the gym environment
                 p_action_space: MSpace = None,  # ptional external action space object that meets the
                 # state space of the gym environment
                 p_logging=Log.C_LOG_ALL):  # Log level (see constants of class Log)

        self._gym_env = p_gym_env
        self.C_NAME = 'Env "' + self._gym_env.spec.id + '"'

        super().__init__(p_mode=Environment.C_MODE_SIM, p_latency=None, p_logging=p_logging)

        if p_state_space is not None:
            self._state_space = p_state_space
        else:
            self._state_space = self.recognize_space(self._gym_env.observation_space)

        if p_action_space is not None:
            self._action_space = p_action_space
        else:
            self._action_space = self.recognize_space(self._gym_env.action_space)

    ## -------------------------------------------------------------------------------------------------
    def __del__(self):
        try:
            self._gym_env.close()
            self.log(self.C_LOG_TYPE_I, 'Closed')
        except:
            pass

    ## -------------------------------------------------------------------------------------------------
    @staticmethod
    def recognize_space(p_gym_space) -> ESpace:
        space = ESpace()

        if isinstance(p_gym_space, gym.spaces.Discrete):
            space.add_dim(
                Dimension(p_name_short='0', p_base_set=Dimension.C_BASE_SET_Z, p_boundaries=[p_gym_space.n]))
        elif isinstance(p_gym_space, gym.spaces.Box):
            shape_dim = len(p_gym_space.shape)
            for i in range(shape_dim):
                for d in range(p_gym_space.shape[i]):
                    space.add_dim(Dimension(p_name_short=str(d), p_base_set=Dimension.C_BASE_SET_R,
                                            p_boundaries=[p_gym_space.low[d], p_gym_space.high[d]]))

        return space

    ## -------------------------------------------------------------------------------------------------
    @staticmethod
    def setup_spaces():
        return None, None

    ## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed=None):

        # 1 Reset Gym environment and determine initial state
        try:
            observation = self._gym_env.reset(seed=p_seed)
        except:
           self._gym_env.seed(p_seed)
           observation = self._gym_env.reset() 
        obs = DataObject(observation)

        # 2 Create state object from Gym observation
        state = State(self._state_space)
        state.set_values(obs.get_data())
        self._set_state(state)

    ## -------------------------------------------------------------------------------------------------
    def simulate_reaction(self, p_state: State, p_action: Action) -> State:

        # 1 Convert action to Gym syntax
        action_sorted = p_action.get_sorted_values()
        dtype = self._gym_env.action_space.dtype

        if (dtype == np.int32) or (dtype == np.int64):
            action_sorted = action_sorted.round(0)

        if action_sorted.size == 1:
            action_gym = action_sorted.astype(self._gym_env.action_space.dtype)[0]
        else:
            action_gym = action_sorted.astype(self._gym_env.action_space.dtype)

        # 2 Process step of Gym environment
        try:
            observation, reward_gym, done, info = self._gym_env.step(action_gym)
        except:
            observation, reward_gym, done, info = self._gym_env.step(np.atleast_1d(action_gym))

        obs = DataObject(observation)

        # 3 Create state object from Gym observation
        state = State(self._state_space, p_terminal=done)
        state.set_values(obs.get_data())

        # 4 Create reward object
        self._last_reward = Reward(Reward.C_TYPE_OVERALL)
        self._last_reward.set_overall_reward(reward_gym)

        # 5 Return next state
        return state

    ## -------------------------------------------------------------------------------------------------
    def compute_reward(self, p_state_old: State = None, p_state_new: State = None) -> Reward:
        if (p_state_old is not None) or (p_state_new is not None):
            raise NotImplementedError

        return self._last_reward

    ## -------------------------------------------------------------------------------------------------
    def compute_success(self, p_state: State) -> bool:
        return self.get_success()

    ## -------------------------------------------------------------------------------------------------
    def compute_broken(self, p_state: State) -> bool:
        return self.get_broken()

    ## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        self._gym_env.render()

    ## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        self._gym_env.render()

    ## -------------------------------------------------------------------------------------------------
    def get_cycle_limit(self):
        return self._gym_env._max_episode_steps


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class WrEnvMLPro2GYM(gym.Env):
    """
    This class is a ready to use wrapper class for MLPro to OpenAI Gym environments. 
    Objects of this type can be treated as an gym.Env object. Encapsulated 
    MLPro environment must be compatible to class Environment.
    """

    C_TYPE = 'MLPro to Gym Env'
    metadata = {'render.modes': ['human']}

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_mlpro_env, p_state_space: MSpace = None, p_action_space: MSpace = None):
        """
        Parameters:
            p_mlpro_env     MLPro's Environment object
            p_state_space   Optional external state space object that meets the
                            state space of the MLPro environment
            p_action_space  Optional external action space object that meets the
                            state space of the MLPro environment
        """

        self._mlpro_env = p_mlpro_env

        if p_state_space is not None:
            self.observation_space = p_state_space
        else:
            self.observation_space = self.recognize_space(self._mlpro_env.get_state_space())

        if p_action_space is not None:
            self.action_space = p_action_space
        else:
            self.action_space = self.recognize_space(self._mlpro_env.get_action_space())

        self.first_refresh = True

    ## -------------------------------------------------------------------------------------------------
    @staticmethod
    def recognize_space(p_mlpro_space):
        space = None
        action_dim = p_mlpro_space.get_num_dim()
        id_dim = p_mlpro_space.get_dim_ids()[0]
        if len(p_mlpro_space.get_dim(id_dim).get_boundaries()) == 1:
            space = gym.spaces.Discrete(p_mlpro_space.get_dim(id_dim).get_boundaries()[0])
        else:
            lows = []
            highs = []
            for dimension in range(action_dim):
                id_dim = p_mlpro_space.get_dim_ids()[dimension]
                lows.append(p_mlpro_space.get_dim(id_dim).get_boundaries()[0])
                highs.append(p_mlpro_space.get_dim(id_dim).get_boundaries()[1])

            space = gym.spaces.Box(
                low=np.array(lows, dtype=np.float32),
                high=np.array(highs, dtype=np.float32),
                shape=(action_dim,),
                dtype=np.float32
            )

        return space

    ## -------------------------------------------------------------------------------------------------
    def step(self, action):
        _action = Action()
        _act_set = Set()
        idx = self._mlpro_env._action_space.get_num_dim()

        if isinstance(self.observation_space, gym.spaces.Discrete):
            action = np.array([action])

        for i in range(idx):
            _act_set.add_dim(Dimension('action_' + str(i)))
        _act_elem = Element(_act_set)
        for i in range(idx):
            _ids = _act_elem.get_dim_ids()
            _act_elem.set_value(_ids[i], action[i].item())
        _action.add_elem('0', _act_elem)

        self._mlpro_env.process_action(_action)
        reward = self._mlpro_env.compute_reward()

        obs = None
        if isinstance(self.observation_space, gym.spaces.Box):
            obs = np.array(self._mlpro_env.get_state().get_values(), dtype=np.float32)
        else:
            obs = np.array(self._mlpro_env.get_state().get_values())

        state = self._mlpro_env.get_state()
        done = state.get_terminal()

        info = {}
        info["TimeLimit.truncated"] = state.get_timeout()

        return obs, reward.get_overall_reward(), done, info

    ## -------------------------------------------------------------------------------------------------
    def reset(self, seed=None, options=None):
        self._mlpro_env.reset(seed)
        obs = None
        if isinstance(self.observation_space, gym.spaces.Box):
            obs = np.array(self._mlpro_env.get_state().get_values(), dtype=np.float32)
        else:
            obs = np.array(self._mlpro_env.get_state().get_values())
        return obs

    ## -------------------------------------------------------------------------------------------------
    def render(self, mode='human'):
        try:
            if self.first_refresh:
                self._mlpro_env.init_plot()
                self.first_refresh = False
            else:
                self._mlpro_env.update_plot()
            return True
        except:
            return False

    ## -------------------------------------------------------------------------------------------------
    def close(self):
        self._mlpro_env.__del__()


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl
## -- Module  : models.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-10-26  1.0.0     DA       Splitted modules for sar-, env-, agent- and training classes.
## --                                Further change logs are documented there.
## -------------------------------------------------------------------------------------------------

"""
This module provides model classes for reinforcement learning tasks. See sub-mdules for further
information.
"""

from mlpro.rl.models_sar import *
from mlpro.rl.models_env import *
from mlpro.rl.models_agents import *
from mlpro.rl.models_train import *



#!/usr/bin/env python

# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Robotiq, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Robotiq, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Copyright (c) 2012, Robotiq, Inc.
# Revision $Id$
#
# Modifed from the orginal comModbusTcp by Kelsey Hawkins @ Georgia Tech



"""@package docstring
Module comModbusRtu: defines a class which communicates with Robotiq Grippers using the Modbus RTU protocol. 

The module depends on pymodbus (http://code.google.com/p/pymodbus/) for the Modbus RTU client.
"""

from pymodbus.client.sync import ModbusSerialClient
from math import ceil

class communication():	

   def __init__(self):
      self.client = None
      
   def connectToDevice(self, device):
      """Connection to the client - the method takes the IP address (as a string, e.g. '192.168.1.11') as an argument."""
      self.client = ModbusSerialClient(method='rtu',port=device,stopbits=1, bytesize=8, baudrate=115200, timeout=0.2)
      if not self.client.connect():
          print("Unable to connect to %s" % device)
          return False
      return True

   def disconnectFromDevice(self):
      """Close connection"""
      self.client.close()

   def sendCommand(self, data):   
      """Send a command to the Gripper - the method takes a list of uint8 as an argument. The meaning of each variable depends on the Gripper model (see support.robotiq.com for more details)"""
      #make sure data has an even number of elements   
      if(len(data) % 2 == 1):
         data.append(0)

      #Initiate message as an empty list
      message = []

      #Fill message by combining two bytes in one register
      for i in range(0, len(data)//2):
         message.append((data[2*i] << 8) + data[2*i+1])

      #To do!: Implement try/except 
      try:
         self.client.write_registers(0x03E8, message, unit=0x0009)
      except:
         print("Modbus write operation failure")
         return False
      return True

   def getStatus(self, numBytes):
      """Sends a request to read, wait for the response and returns the Gripper status. The method gets the number of bytes to read as an argument"""
      numRegs = int(ceil(numBytes/2.0))

      #To do!: Implement try/except 
      #Get status from the device
      try:
        response = self.client.read_holding_registers(0x07D0, numRegs, unit=0x0009)
      except Exception as e:
        print(e)
        return None

      # When reading failes, response is of type None 
      if response is None:
      #   print("Failed to receive status")
        return None

      #Instantiate output as an empty list
      output = []

      #Fill the output with the bytes in the appropriate order
      for i in range(0, numRegs):
         output.append((response.getRegister(i) & 0xFF00) >> 8)
         output.append( response.getRegister(i) & 0x00FF)
      
      #Output the result
      return output



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.dsm.sciui
## -- Module  : scenario_dsm1.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-20  0.0.0     DA       Creation
## -- 2021-07-03  1.0.0     DA       Release of first version
## -- 2022-01-06  1.1.0     DA       Integration in mlpro
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.0 (2022-01-06)

Demo scenarios for SciUI framework that shows the reuse of the interactive 2D/3D input space class.
Can be executed directly...
"""



from mlpro.bf.ui.sciui.framework import *
from mlpro.oa.sciui.iis import InteractiveInputSpace
from mlpro.bf.math import *




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class DSM2D(SciUIScenario):

    C_NAME          = 'Data Stream Mining - Processing of 2D Streams'
    C_VERSION       = '0.0.0'
    C_RELEASED      = True
    C_VISIBLE       = True

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        # 1 Add scenario-specific variables to shared db
        InteractiveInputSpace.enrich_shared_db(self.shared_db)
        self.shared_db.iis_ispace.add_dim( Dimension( p_id=0, 
                                                      p_name_short='x1', 
                                                      p_description='', 
                                                      p_name_latex='x_1', 
                                                      p_unit='m', 
                                                      p_unit_latex='m', 
                                                      p_boundaries=[-5,5]) )
        self.shared_db.iis_ispace.add_dim( Dimension( p_id=1,  
                                                      p_name_short='x2', 
                                                      p_description='', 
                                                      p_name_latex='x_2', 
                                                      p_unit='m/s', 
                                                      p_unit_latex='\\frac{m}{s}', 
                                                      p_boundaries=[-25,25]) )

        # 2 Build scenario structure
        self.add_component(InteractiveInputSpace(self.shared_db, p_row=0, p_col=0, p_padx=5, p_logging=self._level))





## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class DSM3D(SciUIScenario):

    C_NAME          = 'Data Stream Mining - Processing of 3D Streams'
    C_VERSION       = '0.0.0'
    C_RELEASED      = True
    C_VISIBLE       = True

## -------------------------------------------------------------------------------------------------
    def init_component(self):
        super().init_component()

        # 1 Add scenario-specific variables to shared db
        InteractiveInputSpace.enrich_shared_db(self.shared_db)
        self.shared_db.iis_ispace.add_dim( Dimension( p_id=0, 
                                                      p_name_short='x1', 
                                                      p_description='', 
                                                      p_name_latex='x_1', 
                                                      p_unit='m', 
                                                      p_unit_latex='m', 
                                                      p_boundaries=[-5,5]) )
        self.shared_db.iis_ispace.add_dim( Dimension( p_id=1, 
                                                      p_name_short='x2', 
                                                      p_description='', 
                                                      p_name_latex='x_2', 
                                                      p_unit='m/s', 
                                                      p_unit_latex='\\frac{m}{s}', 
                                                      p_boundaries=[-25,25]))
        self.shared_db.iis_ispace.add_dim( Dimension( p_id=2, 
                                                      p_name_short='x3', 
                                                      p_description='', 
                                                      p_name_latex='x_3', 
                                                      p_unit='m/s^2', 
                                                      p_unit_latex='\\frac{m}{s^2}', 
                                                      p_boundaries=[-15,15]) )

        # 2 Build scenario structure
        self.add_component(InteractiveInputSpace(self.shared_db, p_row=0, p_col=0, p_padx=5, p_logging=self._level))



        

if (__name__ == '__main__'): 
    from mlpro.bf.ui.sciui.main import SciUI
    SciUI()



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_rl_002_run_agent_with_own_policy_with_gym_environment.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-05-09  0.0.0     DA       Creation
## -- 2021-06-06  1.0.0     DA       Released first version
## -- 2021-08-28  1.1.0     DA       Introduced Policy
## -- 2021-09-11  1.1.0     MRD      Change Header information to match our new library name
## -- 2021-09-29  1.1.1     SY       Change name: WrEnvGym to WrEnvGYM2MLPro
## -- 2021-10-06  1.1.2     DA       Refactoring 
## -- 2021-10-18  1.1.3     DA       Refactoring 
## -- 2021-11-15  1.2.0     DA       Refactoring 
## -- 2021-11-16  1.2.1     DA       Added explicit scenario reset with constant seeding 
## -- 2021-12-03  1.2.2     DA       Refactoring 
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.2.2 (2021-12-03)

This module shows how to run an own policy inside the standard agent model with an OpenAI Gym environment using 
the fhswf_at_ml framework.
"""


from mlpro.bf.math import *
from mlpro.rl.models import *
from mlpro.wrappers.openai_gym import WrEnvGYM2MLPro
import gym
import random




# 1 Implement your own agent policy
class MyPolicy (Policy):

    C_NAME      = 'MyPolicy'

    def set_random_seed(self, p_seed=None):
        random.seed(p_seed)


    def compute_action(self, p_state: State) -> Action:
        # 1 Create a numpy array for your action values 
        my_action_values = np.zeros(self._action_space.get_num_dim())

        # 2 Computing action values is up to you...
        for d in range(self._action_space.get_num_dim()):
            my_action_values[d] = random.random() 

        # 3 Return an action object with your values
        return Action(self._id, self._action_space, my_action_values)


    def _adapt(self, *p_args) -> bool:
        # 1 Adapting the internal policy is up to you...
        self.log(self.C_LOG_TYPE_W, 'Sorry, I am a stupid agent...')

        # 2 Only return True if something has been adapted...
        return False




# 2 Implement your own RL scenario
class MyScenario (RLScenario):

    C_NAME      = 'Matrix'

    def _setup(self, p_mode, p_ada, p_logging):
        # 1 Setup environment
        gym_env     = gym.make('CartPole-v1')
        self._env   = WrEnvGYM2MLPro(gym_env, p_logging=p_logging) 

        # 2 Setup standard single-agent with own policy
        return Agent( p_policy=MyPolicy( p_observation_space=self._env.get_state_space(),
                                         p_action_space=self._env.get_action_space(),
                                         p_buffer_size=1,
                                         p_ada=p_ada,
                                         p_logging=p_logging),    
                      p_envmodel=None,
                      p_name='Smith',
                      p_ada=p_ada,
                      p_logging=p_logging)




# 3 Create scenario and run some cycles
if __name__ == "__main__":
    # 3.1 Parameters for demo mode
    cycle_limit = 100
    logging     = Log.C_LOG_ALL
    visualize   = True
  
else:
    # 3.2 Parameters for internal unit test
    cycle_limit = 10
    logging     = Log.C_LOG_NOTHING
    visualize   = False
 

# 3.3 Create your scenario and run some cycles
myscenario  = MyScenario(
        p_mode=Mode.C_MODE_SIM,
        p_ada=True,
        p_cycle_limit=cycle_limit,
        p_visualize=visualize,
        p_logging=logging
)

myscenario.reset(p_seed=3)
myscenario.run() 


## -------------------------------------------------------------------------------------------------
## -- Project : FH-SWF Automation Technology - Common Code Base (CCB)
## -- Package : mlpro.pool.policies
## -- Module  : randomgenerator
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2022-05-19  0.0.0     SY       Creation
## -- 2022-05-19  1.0.0     SY       Release of first version
## -- 2022-05-20  1.0.1     SY       Remove constructor and raise error for undefined boundaries
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.1 (2022-05-20)

This module providew random genarator for multi purposes, e.g. testing environment, etc..
"""

from mlpro.rl.models import *
import random
         
        
        
## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RandomGenerator(Policy):
    """
    A random policy that generates random actions for each dimension
    of the underlying action space.

    Parameters
    ----------
    p_observation_space : MSpace     
        Subspace of an environment that is observed by the policy.
    p_action_space : MSpace
        Action space object.
    p_buffer_size : int           
        Size of internal buffer. Default = 1.
    p_ada : bool               
        Boolean switch for adaptivity. Default = True.
    p_logging
        Log level (see constants of class Log). Default = Log.C_LOG_ALL.
    """

    C_NAME      = 'RandomGenerator'

## -------------------------------------------------------------------------------------------------
    def compute_action(self, p_state: State) -> Action:
        # 1 Create an empty numpy array
        my_action_values = np.zeros(self._action_space.get_num_dim())

        # 2 Generating random actions based on the dimension of ation space
        ids = self._action_space.get_dim_ids()
        for d in range(self._action_space.get_num_dim()):
            try:
                base_set = self._action_space.get_dim(ids[d]).get_base_set()
            except:
                raise ParamError('Mandatory base set is not defined.')
                
            try:
                if len(self._action_space.get_dim(ids[d]).get_boundaries()) == 1:
                    lower_boundaries = 0
                    upper_boundaries = self._action_space.get_dim(ids[d]).get_boundaries()[0]
                else:
                    lower_boundaries = self._action_space.get_dim(ids[d]).get_boundaries()[0]
                    upper_boundaries = self._action_space.get_dim(ids[d]).get_boundaries()[1]
                if base_set == 'Z' and base_set == 'N':
                    my_action_values[d] = random.randint(lower_boundaries, upper_boundaries)
                elif base_set == 'R' or base_set == 'DO':
                    my_action_values[d] = random.uniform(lower_boundaries, upper_boundaries)
            except:
                raise ParamError('Mandatory boundaries are not defined.')

        # 3 Return an action object with the generated random values
        return Action(self._id, self._action_space, my_action_values)

## -------------------------------------------------------------------------------------------------
    def _adapt(self, *p_args) -> bool:
        self.log(self.C_LOG_TYPE_W, 'Sorry, I am not adapting anything!')
        return False


#!/usr/bin/env python
import rosparam
import rospkg
import os

def LoadYamlFileParamsTest(rospackage_name, rel_path_from_package_to_file, yaml_file_name):

    rospack = rospkg.RosPack()
    pkg_path = rospack.get_path(rospackage_name)
    config_dir = os.path.join(pkg_path, rel_path_from_package_to_file) 
    path_config_file = os.path.join(config_dir, yaml_file_name)
    
    paramlist=rosparam.load_file(path_config_file)
    
    for params, ns in paramlist:
        rosparam.upload_params(ns,params)
        


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : multigeorobot.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-19  0.0.0     MRD      Creation
## -- 2021-12-19  1.0.0     MRD      Released first version
## -- 2022-02-25  1.0.1     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-04-29  1.0.2     MRD      Wrap the environment with WrEnvGYM2MLPro
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.2 (2022-04-29)

This module provides an environment for multi geometry robot.
"""

from mlpro.rl.models import *
import mlpro
import rospy
import subprocess
from mlpro.wrappers.openai_gym import WrEnvGYM2MLPro

from openai_ros.openai_ros_common import StartOpenAI_ROS_Environment
from openai_ros.task_envs.task_commons import LoadYamlFileParamsTest

class MultiGeo(WrEnvGYM2MLPro):
    """
    This module provides an environment for multi geometry robot.
    """

    C_NAME      = 'MultiGeo'
    C_LATENCY   = timedelta(0,5,0)
    C_INFINITY  = np.finfo(np.float32).max  

    def __init__(self, p_seed=0, p_logging=True):
        roscore = subprocess.Popen('roscore')
        rospy.init_node('multi_geo_robot_training', anonymous=True, log_level=rospy.WARN)

        LoadYamlFileParamsTest(rospackage_name="multi_geo_robot_training",
                               rel_path_from_package_to_file="config",
                               yaml_file_name="multi_geo_robot.yaml")  

        ros_ws_path = mlpro.rl.pool.envs.multigeorobot.__file__.replace("/__init__.py", "")
        rospy.set_param('ros_ws_path', ros_ws_path)
        # Init OpenAI_ROS ENV
        task_and_robot_environment_name = rospy.get_param(
        '/multi_geo_robot/task_and_robot_environment_name')

        max_step_episode = rospy.get_param(
        '/multi_geo_robot/max_iterations')  

        env = StartOpenAI_ROS_Environment(task_and_robot_environment_name, max_step_episode)
        env.seed(p_seed)       
        
        super().__init__(p_gym_env=env)
    
## -------------------------------------------------------------------------------------------------
    def compute_success(self, p_state: State) -> bool:
        obs = p_state.get_values()
        close = np.allclose(a=obs[:3],
                            b=obs[3:],
                            atol=0.1)

        if close:
            self._state.set_terminal(True)

        return close


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_bf_006_buffers.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-10-26  1.0.0     SY       Creation/Release
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.0 (2021-10-26)

This module demonstrates how to use classes Buffer and BufferElement.
"""


from mlpro.bf.data import *
import random



if __name__ == "__main__":

    # 1. Instantiate a buffer with random sampling
    buffer     = BufferRnd(p_size=100)
    num_cycles = 150


    # 2 Generate random values and store them to the Buffer
    for i in range(num_cycles):
        
        # 2.1 Store the values and their names in a BufferElement
        buffer_element  =  BufferElement({"reward":random.uniform(-10,10),
                                        "actions":[random.uniform(0,1),random.uniform(0,1)]})
        
        # 2.2 Example: add value element in the developed BufferElement
        buffer_element.add_value_element(dict(accuracy=random.uniform(0,1)))
        
        # 2.3 Add the BufferElement into the Buffer
        buffer.add_element(buffer_element)
        print('Cycle : %.i'%int(i+1))
        
        # 2.4 Checking whether buffer is full or not
        if not buffer.is_full():
            print('Buffer is not full yet, keep collecting data!\n')
        else:
            print('Buffer is full, ready to use!')
            
            # 2.5 Get all data from the Buffer
            all_data = buffer.get_all()
            _actions            = all_data["actions"]
            _reward             = all_data["reward"]
            _accuracy           = all_data["accuracy"]
            
            # 2.6 Get sample data from the Buffer, you define your sampling strategy by
            # redifining method _gen_sample_ind(self, p_num:int)
            sample_data = buffer.get_sample(p_num=10)
            print('Get sample!\n')
            _actions_sample    = sample_data["actions"]
            _reward_sample     = sample_data["reward"]
            _accuracy_sample   = sample_data["accuracy"]

    # 3 To clear your buffer        
    if buffer is not None:
        buffer.clear()
        print('Buffer is cleared!')




#!/usr/bin/env python
from gym import utils
import copy
import math
import rospy
import rospkg
from gym import spaces
from openai_ros.robot_envs import ur5_lab_env
import numpy as np
from sensor_msgs.msg import JointState
from openai_ros.openai_ros_common import ROSLauncher
from openai_ros.task_envs.task_commons import LoadYamlFileParamsTest
import os

class UR5LabRealTask(ur5_lab_env.UR5LabRealEnv, utils.EzPickle):
    def __init__(self):
        # Load Params from the desired Yaml file relative to this TaskEnvironment
        LoadYamlFileParamsTest(rospackage_name="openai_ros",
                               rel_path_from_package_to_file="src/openai_ros/task_envs/ur5_lab/config",
                               yaml_file_name="simple_task.yaml")

        # This is the path where the simulation files are

        ros_ws_abspath = rospy.get_param("ros_ws_path", None)
        visualize = rospy.get_param("visualize", False)
        sim_mode = rospy.get_param("sim", True)
        assert ros_ws_abspath is not None, "You forgot to set ros_ws_abspath in your yaml file of your main RL script. Set ros_ws_abspath: \'YOUR/SIM_WS/PATH\'"
        assert os.path.exists(ros_ws_abspath), "The Simulation ROS Workspace path " + ros_ws_abspath + \
                                               " DOESNT exist, execute: mkdir -p " + ros_ws_abspath + \
                                               "/src;cd " + ros_ws_abspath + ";catkin_make"

        ROSLauncher(rospackage_name="ur_gazebo",
                    launch_file_name="ur5_lab_world.launch",
                    launch_arguments=dict(gui=visualize, sim=sim_mode),
                    ros_ws_abspath=ros_ws_abspath)

        super(UR5LabRealTask, self).__init__(ros_ws_abspath)

        rospy.logdebug("Entered UR5LabRealTask Env")
        self.get_params()

        self.action_space = spaces.Box(low=-0.1, high=0.1, shape=(6,))

        self.n_observations = 6
        self.observation_space = spaces.Box(low=-2, high=2, shape=(6,))
        self.init_distance = None

    def get_params(self):
        # get configuration parameters

        init_pos_dict = rospy.get_param('/ur5_lab/init_pos')
        self.init_pos = [init_pos_dict["shoulder_pan_joint"],
                         init_pos_dict["shoulder_lift_joint"],
                         init_pos_dict["elbow_joint"],
                         init_pos_dict["wrist_1_joint"],
                         init_pos_dict["wrist_2_joint"],
                         init_pos_dict["wrist_3_joint"],
                         ]

        goal_pos_dict = rospy.get_param('/ur5_lab/goal_pos')
        self.goal_pos = [goal_pos_dict["x_pos"],
                         goal_pos_dict["y_pos"],
                         goal_pos_dict["z_pos"]]

        self._tcp_offset_position_x = rospy.get_param('/ur5_lab/offset_real/x_pos', 0)
        self._tcp_offset_position_y = rospy.get_param('/ur5_lab/offset_real/y_pos', 0)
        self._tcp_offset_position_z = rospy.get_param('/ur5_lab/offset_real/z_pos', 0)

        self.reached_goal_reward = rospy.get_param(
            '/ur5_lab/reached_goal_reward')

    def _set_init_pose(self):
        """
        Sets the Robot in its init pose
        """
        # Check because it seems its not being used
        rospy.logdebug("Init Pos:")
        rospy.logdebug(self.init_pos)

        # INIT POSE
        rospy.logdebug("Moving To Init Pose ")
        self.move_joints(self.init_pos)
        self.last_action = "INIT"

        pose = self.get_ee_pose()
        pose = np.array([pose.position.x,
                         pose.position.y,
                         pose.position.z, ])

        self.init_distance = np.linalg.norm(np.array(pose[:]) - np.array(self.goal_pos[:]))

        return True

    def _init_env_variables(self):
        rospy.logdebug("Init Env Variables...")
        rospy.logdebug("Init Env Variables...END")

    def _set_action(self, action):

        gripper_target = self.get_joint_states()
        gripper_target = gripper_target + action
        gripper_target = np.clip(gripper_target, -math.pi, math.pi)
        self.last_action = "Joint Move"

        self.movement_result = self.move_joints(gripper_target)

        rospy.logwarn("END Set Action ==>" + str(action) +
                      ", NAME=" + str(self.last_action))

    def _get_obs(self):
        current_pose = self.get_ee_pose()
        self._tcp_pose_pub.publish(current_pose)
        self._tcp_real_msg.position.x = self._tcp_real_msg.position.x + self._tcp_offset_position_x
        self._tcp_real_msg.position.y = self._tcp_real_msg.position.y + self._tcp_offset_position_y
        self._tcp_real_msg.position.z = self._tcp_real_msg.position.z + self._tcp_offset_position_z
        self._tcp_real_pose_pub.publish(self._tcp_real_msg)
        obs_pose = np.array([current_pose.position.x,
                             current_pose.position.y,
                             current_pose.position.z,
                             self.goal_pos[0],
                             self.goal_pos[1],
                             self.goal_pos[2], ])

        rospy.logdebug("OBSERVATIONS====>>>>>>>" + str(obs_pose))

        return obs_pose

    def _is_done(self, observations):
        done = np.allclose(a=observations[:3],
                           b=observations[3:],
                           atol=0.1)

        return done or self.no_motion_plan

    def _compute_reward(self, observations, done):
        """
        We punish each step that it passes without achieveing the goal.
        Punishes differently if it reached a position that is imposible to move to.
        Rewards getting to a position close to the goal.
        """

        distance = np.linalg.norm(np.array(observations[:3]) - np.array(observations[3:]))
        ratio = distance / self.init_distance
        reward = -np.ones(1) * ratio
        reward = reward - 10e-2

        if done:
            if self.no_motion_plan:
                reward += -(self.reached_goal_reward * 4)
            else:
                reward += self.reached_goal_reward

        rospy.logwarn(">>>REWARD>>>" + str(reward))

        return reward



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : robotinhtm
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-09-10  0.0.0     MRD      Creation
## -- 2021-09-11  1.0.0     MRD      Release of first version
## -- 2021-09-11  1.0.1     MRD      Change Header information to match our new library name
## -- 2021-09-25  1.0.2     MRD      Minor fix for state space and action space recognition
## -- 2021-10-05  1.0.3     SY       Update following new attributes done and broken in State
## -- 2021-11-15  1.1.0     DA       Refactoring of class RobotHTM
## -- 2021-12-03  1.1.1     DA       Refactoring
## -- 2021-12-08  1.1.2     MRD      Change the state, include the joint angles
## -- 2021-12-19  1.1.3     DA       Replaced 'done' by 'success'
## -- 2021-12-21  1.1.4     DA       Class RobotHTM: renamed method reset() to _reset()
## -- 2021-12-21  1.1.5     MRD      Add Termination on Success
## -- 2022-01-21  1.1.6     MRD      Add recommended cycle limit and add seed parameter
## -- 2022-02-25  1.1.7     SY       Refactoring due to auto generated ID in class Dimension
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.1.7 (2022-02-25)

This module provides an environment of a robot manipulator based on Homogeneous Matrix
"""
import torch
import transformations
from mlpro.rl.models import *


## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class RobotArm3D:
    """
    Auxiliary class for the implementation of robotinhtm.
    Generate the Kinematic of a pre-defined robot in Homogeneous Matrix.
    """

    def __init__(self):
        self.thetas = torch.Tensor([])
        self.joints = torch.Tensor([])
        self.orientation = torch.Tensor([])
        self.adjustRot = []
        self.adjustTheta = []
        self.num_joint = 0
        self.jointAxis = []
        self.lvector = torch.Tensor([])
        self.HM = torch.Tensor([])
        self.HMeef = None

    ## -------------------------------------------------------------------------------------------------
    def add_link_joint(
            self,
            jointAxis=None,
            lvector=None,
            thetaInit=None,
            adjustRot=None,
            adjustTheta=None,
    ):
        self.joints = torch.cat([self.joints, torch.Tensor([[0, 0, 0, 1]]).T], dim=1)
        self.jointAxis.append(jointAxis)
        self.lvector = torch.cat([self.lvector, lvector])
        if adjustRot is not None:
            self.adjustRot.append(adjustRot)
            self.adjustTheta.append(adjustTheta)
        else:
            self.adjustRot.append(None)
            self.adjustTheta.append(None)
        if jointAxis != "f":
            self.thetas = torch.cat([self.thetas, thetaInit])
            self.num_joint = self.num_joint + 1

    ## -------------------------------------------------------------------------------------------------
    def get_transformation_matrix(self, theta, lvector, rotAxis, adjustRots=None, adjustThetas=None):
        transformationMatrix = torch.Tensor([])

        if rotAxis == "rx":
            rotM = torch.Tensor(
                [
                    [1, 0, 0],
                    [0, torch.cos(theta), -torch.sin(theta)],
                    [0, torch.sin(theta), torch.cos(theta)],
                ]
            )
        elif rotAxis == "ry":
            rotM = torch.Tensor(
                [
                    [torch.cos(theta), 0, torch.sin(theta)],
                    [0, 1, 0],
                    [-torch.sin(theta), 0, torch.cos(theta)],
                ]
            )
        elif rotAxis == "rz":
            rotM = torch.Tensor(
                [
                    [torch.cos(theta), -torch.sin(theta), 0],
                    [torch.sin(theta), torch.cos(theta), 0],
                    [0, 0, 1],
                ]
            )
        elif rotAxis == "f":
            rotM = torch.eye(3)

        # Adjust Rotation Projection
        if adjustRots is None:
            rotM = torch.matmul(rotM, torch.eye(3))
        else:
            rotMsT = torch.eye(3)
            for projection in range(len(adjustRots)):
                if adjustRots[projection] == "rx":
                    rotMs = torch.Tensor(
                        [
                            [1, 0, 0],
                            [
                                0,
                                torch.cos(adjustThetas[projection]),
                                -torch.sin(adjustThetas[projection]),
                            ],
                            [
                                0,
                                torch.sin(adjustThetas[projection]),
                                torch.cos(adjustThetas[projection]),
                            ],
                        ]
                    )
                elif adjustRots[projection] == "ry":
                    rotMs = torch.Tensor(
                        [
                            [
                                torch.cos(adjustThetas[projection]),
                                0,
                                torch.sin(adjustThetas[projection]),
                            ],
                            [0, 1, 0],
                            [
                                -torch.sin(adjustThetas[projection]),
                                0,
                                torch.cos(adjustThetas[projection]),
                            ],
                        ]
                    )
                elif adjustRots[projection] == "rz":
                    rotMs = torch.Tensor(
                        [
                            [
                                torch.cos(adjustThetas[projection]),
                                -torch.sin(adjustThetas[projection]),
                                0,
                            ],
                            [
                                torch.sin(adjustThetas[projection]),
                                torch.cos(adjustThetas[projection]),
                                0,
                            ],
                            [0, 0, 1],
                        ]
                    )
                rotMsT = torch.mm(rotMsT, rotMs)
            rotM = torch.mm(rotM, rotMsT)

        transformationMatrix = torch.cat([rotM, lvector.reshape(1, 3).T], dim=1)
        transformationMatrix = torch.cat([transformationMatrix, torch.Tensor([[0, 0, 0, 1]])], dim=0)
        return transformationMatrix

    ## -------------------------------------------------------------------------------------------------
    def update_joint_coords(self):
        self.HM = torch.Tensor([])
        T = torch.eye(4)
        for i in range(len(self.lvector)):
            if self.jointAxis[i] != "f":
                T_next = self.get_transformation_matrix(
                    self.thetas[i],
                    self.lvector[i],
                    rotAxis=self.jointAxis[i],
                    adjustRots=self.adjustRot[i],
                    adjustThetas=self.adjustTheta[i],
                )
            else:
                T_next = self.get_transformation_matrix(
                    torch.zeros(1),
                    self.lvector[i],
                    rotAxis=self.jointAxis[i],
                    adjustRots=self.adjustRot[i],
                    adjustThetas=self.adjustTheta[i],
                )

            T = torch.mm(T, T_next)
            self.HM = torch.cat([self.HM, T])
            self.joints[:, [i]] = torch.mm(T, torch.Tensor([[0, 0, 0, 1]]).T)

        # self.orientation = self.convert_to_quaternion_only()
        self.HMeef = T

    ## -------------------------------------------------------------------------------------------------
    def get_joint(self):
        return self.joints

    ## -------------------------------------------------------------------------------------------------
    def get_num_joint(self):
        return self.num_joint

    ## -------------------------------------------------------------------------------------------------
    def set_theta(self, theta):
        self.thetas = theta.flatten()

    ## -------------------------------------------------------------------------------------------------
    def get_homogeneous(self):
        return self.HM

    ## -------------------------------------------------------------------------------------------------
    def get_homogeneous_eef(self):
        return self.HMeef

    ## -------------------------------------------------------------------------------------------------
    def get_orientation(self):
        return self.orientation

    ## -------------------------------------------------------------------------------------------------
    def update_theta(self, deltaTheta):
        self.thetas += deltaTheta.flatten()

    ## -------------------------------------------------------------------------------------------------
    def convert_to_quaternion(self):
        origin, xaxis, yaxis, zaxis = [0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]
        hm_reshape = self.HM.reshape(self.joints.shape[1], 4, 4)
        hm_length = hm_reshape.shape[0]

        out = torch.Tensor([])
        for x in range(hm_length):
            _, _, rot, trans, _ = transformations.decompose_matrix(hm_reshape[x].numpy())

            qx = transformations.quaternion_about_axis(rot[0], xaxis)
            qy = transformations.quaternion_about_axis(rot[1], yaxis)
            qz = transformations.quaternion_about_axis(rot[2], zaxis)
            q = transformations.quaternion_multiply(qx, qy)
            q = transformations.quaternion_multiply(q, qz)

            outs = torch.Tensor(trans)
            outs = torch.cat([outs, torch.Tensor(q)])
            out = torch.cat([out, outs])

        return out

    ## -------------------------------------------------------------------------------------------------


## -------------------------------------------------------------------------------------------------
class RobotHTM(Environment):
    """
    Custom environment for an arm robot represented as Homogeneous Matrix.
    The goal is to reach a certain point.
    """

    C_NAME = "RobotHTM"
    C_REWARD_TYPE = Reward.C_TYPE_OVERALL
    C_LATENCY = timedelta(0, 1, 0)
    C_INFINITY = np.finfo(np.float32).max
    C_CYCLE_LIMIT = 100

    ## -------------------------------------------------------------------------------------------------
    def __init__(self, p_num_joints=4, p_reset_seed=True, p_seed=None, p_target_mode="random", p_logging=Log.C_LOG_ALL):
        """
        Parameters:
            p_logging               Boolean switch for logging
            p_num_joints            Number of joints
            p_target_mode           Target Mode (Random or Fixed)
        """

        self.RobotArm1 = RobotArm3D()

        joints = []

        jointType = []
        vectLinkLength = [[0, 0, 0], [0, 0, 0]]
        jointType.append("rz")
        for joint in range(p_num_joints - 1):
            vectLinkLength.append([0, 0.7, 0])
            jointType.append("rx")

        jointType.append("f")

        for x in range(len(jointType)):
            vectorLink = dict(x=vectLinkLength[x][0], y=vectLinkLength[x][1], z=vectLinkLength[x][2])
            joint = dict(
                Joint_name="Joint %d" % x,
                Joint_type=jointType[x],
                Vector_link_length=vectorLink,
            )
            joints.append(joint)

        for robo in joints:
            self.RobotArm1.add_link_joint(
                lvector=torch.Tensor(
                    [
                        [
                            robo["Vector_link_length"]["x"],
                            robo["Vector_link_length"]["y"],
                            robo["Vector_link_length"]["z"],
                        ]
                    ]
                ),
                jointAxis=robo["Joint_type"],
                thetaInit=torch.Tensor([np.radians(0)]),
            )

        self.RobotArm1.update_joint_coords()
        self.jointangles = self.RobotArm1.thetas
        self.dt = 0.01
        self.target_mode = p_target_mode
        self.target = None
        self.init_distance = None
        self.num_joint = self.RobotArm1.get_num_joint()
        self.reach = torch.norm(torch.Tensor([[0.0, 0.0, 0.0]]) - self.RobotArm1.joints[:3, [-1]].reshape(1, 3))
        self.last_distance = None

        super().__init__(p_mode=Environment.C_MODE_SIM, p_logging=p_logging)
        self._state_space, self._action_space = self._setup_spaces()
        self.set_random_seed(p_seed)
        self._reset_seed = p_reset_seed

        self.reset()

    ## -------------------------------------------------------------------------------------------------
    @staticmethod
    def setup_spaces():
        return None, None

    ## -------------------------------------------------------------------------------------------------
    def _setup_spaces(self):
        """
        Implement this method to enrich the state and action space with specific
        dimensions.
        """

        # 1 Setup state space
        state_space = ESpace()
        state_space.add_dim(Dimension("Tx", "Targetx", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
        state_space.add_dim(Dimension("Ty", "Targety", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
        state_space.add_dim(Dimension("Tz", "Targetz", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
        state_space.add_dim(Dimension("Px", "Targetx", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
        state_space.add_dim(Dimension("Py", "Targety", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
        state_space.add_dim(Dimension("Pz", "Targetz", "", "m", "m", p_boundaries=[-np.inf, np.inf]))

        for idx in range(self.num_joint):
            state_space.add_dim(
                Dimension(
                    "J%i" % (idx),
                    "Joint%i" % (idx),
                    "",
                    "deg",
                    "deg",
                    p_boundaries=[-np.inf, np.inf],
                )
            )

        # 2 Setup action space
        action_space = ESpace()
        for idx in range(self.num_joint):
            action_space.add_dim(
                Dimension(
                    "A%i" % (idx),
                    "AV%i" % (idx),
                    "",
                    "rad/sec",
                    "\frac{rad}{sec}",
                    p_boundaries=[-np.pi, np.pi],
                )
            )

        return state_space, action_space

    ## -------------------------------------------------------------------------------------------------
    def _simulate_reaction(self, p_state: State, p_action: Action) -> State:
        action = p_action.get_sorted_values()
        if not isinstance(action, torch.Tensor):
            action = torch.Tensor(action)
        self.RobotArm1.update_theta(action * self.dt)
        self.RobotArm1.update_joint_coords()
        self.jointangles = self.RobotArm1.thetas

        obs = torch.cat(
            [
                self.target,
                self.RobotArm1.joints[:3, [-1]].reshape(1, 3),
                self.RobotArm1.thetas.reshape(1, self.num_joint),
            ],
            dim=1,
        )
        obs = obs.cpu().flatten().tolist()
        state = State(self._state_space)
        state.set_values(obs)

        return state

    ## -------------------------------------------------------------------------------------------------
    def _compute_success(self, p_state: State = None) -> bool:
        disterror = np.linalg.norm(np.array(p_state.get_values())[:3] - np.array(p_state.get_values())[3:6])

        if disterror <= 0.1:
            self._state.set_terminal(True)
            return True
        else:
            return False

    ## -------------------------------------------------------------------------------------------------
    def _compute_broken(self, p_state: State) -> bool:
        return False

    ## -------------------------------------------------------------------------------------------------
    def _compute_reward(self, p_state_old: State, p_state_new: State) -> Reward:
        reward = Reward(self.C_REWARD_TYPE)
        disterror = np.linalg.norm(np.array(p_state_new.get_values())[:3] - np.array(p_state_new.get_values())[3:6])

        ratio = disterror / self.init_distance.item()
        rew = -np.ones(1) * ratio
        rew = rew - 10e-2
        if disterror <= 0.1:
            rew = rew + 1
        rew = rew.astype("float64")
        reward.set_overall_reward(rew.item())
        return reward

    ## -------------------------------------------------------------------------------------------------
    def set_theta(self, theta):
        self.RobotArm1.thetas = theta.reshape(self.num_joint)
        self.RobotArm1.update_joint_coords()
        self.jointangles = self.RobotArm1.thetas

    ## -------------------------------------------------------------------------------------------------
    def _reset(self, p_seed=None) -> None:
        if self._reset_seed:
            self.set_random_seed(p_seed)
        theta = torch.zeros(self.RobotArm1.get_num_joint())
        self.RobotArm1.set_theta(theta)
        self.RobotArm1.update_joint_coords()
        self.jointangles = self.RobotArm1.thetas
        if self.target_mode == "random":
            num = random.random()
            if num < 0.2:
                self.target = torch.Tensor([[0.5, 0.5, 0.5]])
                self.init_distance = torch.norm(self.RobotArm1.joints[:3, [-1]].reshape(1, 3) - self.target)
            elif num < 0.4:
                self.target = torch.Tensor([[0.0, 0.5, 0.5]])
                self.init_distance = torch.norm(self.RobotArm1.joints[:3, [-1]].reshape(1, 3) - self.target)
            elif num < 0.6:
                self.target = torch.Tensor([[-0.5, 0.0, 0.5]])
                self.init_distance = torch.norm(self.RobotArm1.joints[:3, [-1]].reshape(1, 3) - self.target)
            elif num < 0.8:
                self.target = torch.Tensor([[0.0, -0.5, 0.5]])
                self.init_distance = torch.norm(self.RobotArm1.joints[:3, [-1]].reshape(1, 3) - self.target)
            else:
                self.target = torch.Tensor([[-0.5, -0.5, 0.5]])
                self.init_distance = torch.norm(self.RobotArm1.joints[:3, [-1]].reshape(1, 3) - self.target)
        else:
            self.target = torch.Tensor([[0.5, 0.5, 0.5]])
            self.init_distance = torch.norm(self.RobotArm1.joints[:3, [-1]].reshape(1, 3) - self.target)
        obs = torch.cat(
            [
                self.target,
                self.RobotArm1.joints[:3, [-1]].reshape(1, 3),
                self.RobotArm1.thetas.reshape(1, self.num_joint),
            ],
            dim=1,
        )
        obs = obs.cpu().flatten().tolist()
        self._state = State(self._state_space)
        self._state.set_values(obs)

    ## -------------------------------------------------------------------------------------------------
    def init_plot(self, p_figure=None):
        pass

    ## -------------------------------------------------------------------------------------------------
    def update_plot(self):
        pass



#!/usr/bin/env python
import serial
from serial.serialutil import SerialException

from pymodbus.client.sync import ModbusSerialClient
import robotiq_modbus_rtu.comModbusRtu

from math import ceil

import numpy as np
import array

ACTION_REQ_IDX = 7
POS_INDEX      = 10 
SPEED_INDEX    = 11
FORCE_INDEX    = 12

class Robotiq2FingerGripper():
    def __init__(self, device_id=0, stroke=0.085, comport='/dev/ttyUSB0',baud=115200):

        self.client = robotiq_modbus_rtu.comModbusRtu.communication()
        
        connected = self.client.connectToDevice(device = comport)
        if not connected:
            raise Exception("Communication with gripper %d on serial port: %s and baud rate: %d not achieved" % (device_id, comport, baud))
        
        self.init_success = True
        self.device_id = device_id+9
        self.stroke = stroke
        self.initialize_communication_variables()

        self.message = []       
    
    def _update_cmd(self):

         #Initiate command as an empty list
        self.message = []
        #Build the command with each output variable
        self.message.append(self.rACT + (self.rGTO << 3) + (self.rATR << 4))
        self.message.append(0)
        self.message.append(0)
        self.message.append(self.rPR)
        self.message.append(self.rSP)
        self.message.append(self.rFR)   

    def sendCommand(self):
        """Send the command to the Gripper."""    
        return self.client.sendCommand(self.message)


    def getStatus(self):
        """Request the status from the gripper and return it in the Robotiq2FGripper_robot_input msg type."""

        #Acquire status from the Gripper
        status = self.client.getStatus(6);

        # Check if read was successful
        if( status is None ):
          return False

        #Assign the values to their respective variables
        self.gACT = (status[0] >> 0) & 0x01;        
        self.gGTO = (status[0] >> 3) & 0x01;
        self.gSTA = (status[0] >> 4) & 0x03;
        self.gOBJ = (status[0] >> 6) & 0x03;
        self.gFLT =  status[2]
        self.gPR  =  status[3]
        self.gPO  =  status[4]
        self.gCU  =  status[5]       

        return True
        

    def initialize_communication_variables(self):
        # Out
        self.rPR = 0
        self.rSP = 255
        self.rFR = 150
        self.rARD = 1
        self.rATR = 0
        self.rGTO = 0
        self.rACT = 0
        # In
        self.gSTA = 0
        self.gACT = 0
        self.gGTO = 0
        self.gOBJ = 0
        self.gFLT = 0
        self.gPO = 0
        self.gPR = 0
        self.gCU = 0

        self._update_cmd()
        self._max_force = 100.0 # [%]
        
    def shutdown(self):
        self.client.close()
    
    def activate_gripper(self):
        self.rACT = 1
        self.rPR = 0
        self.rSP = 255
        self.rFR = 150
        self._update_cmd()
    
    def deactivate_gripper(self):
        self.rACT = 0
        self._update_cmd()
        
    def activate_emergency_release(self,open_gripper=True):
        self.rATR = 1
        self.rARD = 1

        if (open_gripper):
            self.rARD=0
        self._update_cmd()
                
    def deactivate_emergency_release(self):
        self.rATR = 0
        self._update_cmd()

    def goto(self, pos, vel, force):
        self.rACT = 1
        self.rGTO = 1
        self.rPR = int(np.clip((3. - 230.)/self.stroke * pos + 230., 0, 255))
        self.rSP = int(np.clip(255./(0.1 - 0.013) * vel-0.013, 0, 255))
        self.rFR = int(np.clip(255./(self._max_force) * force, 0, 255))
        self._update_cmd()
        

    def stop(self):
        self.rACT = 1
        self.rGTO = 0
        self._update_cmd()
                
    def is_ready(self):
        return self.gSTA == 3 and self.gACT == 1

    def is_reset(self):
        return self.gSTA == 0 or self.gACT == 0

    def is_moving(self):
        return self.gGTO == 1 and self.gOBJ == 0

    def is_stopped(self):
        return self.gOBJ != 0

    def object_detected(self):
        return self.gOBJ == 1 or self.gOBJ == 2

    def get_fault_status(self):
        return self.gFLT

    def get_pos(self):
        po = float(self.gPO)
        return np.clip(self.stroke/(3.-230.)*(po-230.), 0, self.stroke)

    def get_req_pos(self):
        pr = float(self.gPR)
        return np.clip(self.stroke/(3.-230.)*(pr-230.), 0, self.stroke)

    def get_current(self):
        return self.gCU * 0.1

    



#!/usr/bin/env python
from gym.envs.registration import register
from gym import envs
import gym
import rospy

def RegisterOpenAI_Ros_Env(task_env, max_episode_steps=10000):
    """
    Registers all the ENVS supported in OpenAI ROS. This way we can load them
    with variable limits.
    Here is where you have to PLACE YOUR NEW TASK ENV, to be registered and accesible.
    return: False if the Task_Env wasnt registered, True if it was.
    """

    ###########################################################################
    result = True
    if task_env in gym.envs.registry.env_specs:
        rospy.logwarn("Environment already registered, . . .\n re-registering. . .")
        del gym.envs.registry.env_specs[task_env]
        
    if task_env == 'UR5LabSimTask-v0':
        # We have to import the Class that we registered so that it can be found afterwards in the Make
        from openai_ros.task_envs.ur5_lab import sim_task_v0
        # We register the Class through the Gym system
        register(
            id=task_env,
            entry_point='openai_ros.task_envs.ur5_lab.sim_task_v0:UR5LabSimTask',
            max_episode_steps=max_episode_steps,
        )
        
    elif task_env == 'UR5LabRealTask-v0':
        # We have to import the Class that we registered so that it can be found afterwards in the Make
        from openai_ros.task_envs.ur5_lab import real_task_v0
        # We register the Class through the Gym system
        register(
            id=task_env,
            entry_point='openai_ros.task_envs.ur5_lab.real_task_v0:UR5LabRealTask',
            max_episode_steps=max_episode_steps,
        )
        
    else:
        result = False

    ###########################################################################

    if result:
        # We check that it was really registered
        supported_gym_envs = GetAllRegisteredGymEnvs()
        #print("REGISTERED GYM ENVS===>"+str(supported_gym_envs))
        assert (task_env in supported_gym_envs), "The Task_Robot_ENV given is not Registered ==>" + \
            str(task_env)

    return result


def GetAllRegisteredGymEnvs():
    """
    Returns a List of all the registered Envs in the system
    return EX: ['Copy-v0', 'RepeatCopy-v0', 'ReversedAddition-v0', ... ]
    """

    all_envs = envs.registry.all()
    env_ids = [env_spec.id for env_spec in all_envs]

    return env_ids



#!/usr/bin/env python3
# import geometry_msgs.msg
import rospy
import actionlib
from robotiq_2f_gripper_msgs.msg import CommandRobotiqGripperFeedback, CommandRobotiqGripperResult, CommandRobotiqGripperAction, CommandRobotiqGripperGoal
from robotiq_2f_gripper_control.robotiq_2f_gripper_driver import Robotiq2FingerGripperDriver as Robotiq
import control_msgs.msg
from std_msgs.msg import String
from gazebo_msgs.msg import ContactsState
from gazebo_ros_link_attacher.srv import Attach, AttachRequest
import math

rospy.init_node('gripper_command')
sim = rospy.get_param('~sim','false')
if sim:
    print("""
            
            THIS IS A SIMULATION 
            
            """)
else:
    print("""
            
            Using Real Hardware 
            
            """)
if sim:
    client = actionlib.SimpleActionClient('/gripper_controller/gripper_cmd', control_msgs.msg.GripperCommandAction) 
else:
    client = actionlib.SimpleActionClient('/gripper_controller/gripper_cmd', CommandRobotiqGripperAction)

client.wait_for_server()

attach_srv = rospy.ServiceProxy('/link_attacher_node/attach',Attach)
attach_srv.wait_for_service(timeout=5)
rospy.loginfo("Created ServiceProxy to /link_attacher_node/attach")
detach_srv = rospy.ServiceProxy('/link_attacher_node/detach',Attach)
detach_srv.wait_for_service(timeout=5)
rospy.loginfo("Created ServiceProxy to /link_attacher_node/detach")
rospy.loginfo("Gripper Contact Handler Ready")

class Comparator():
    def __init__(self):
        self.process=False
        self.r_contact_col_name=None
        self.r_contact_norm=None
        self.l_contact_col_name=None
        self.l_contact_norm=None
    def r_call(self,contact_array):
        if self.process==False:
            return
        for state in contact_array.states:
            self.r_contact_col_name=(state.collision1_name if state.collision1_name !=
                                     "robot::right_inner_finger::right_inner_finger_pad_collision_1"
                                     else state.collision2_name)
            for c_n in state.contact_normals:
                if (math.isclose(0,c_n.x,abs_tol=0.01) and 
                    math.isclose(1,c_n.y,rel_tol=0.1) and 
                    math.isclose(0,c_n.z,abs_tol=0.01)):
                    self.r_contact_norm=True
                    break
                else:
                    self.r_contact_norm=False
                    continue
            self.cancel_goal()
        return
                
    def l_call(self,contact_array):
        if self.process==False:
            return
        for state in contact_array.states:
            self.l_contact_col_name=(state.collision1_name if state.collision1_name !=
                                     "robot::left_inner_finger::left_inner_finger_pad_collision_1"
                                     else state.collision2_name)
            for c_n in state.contact_normals:
                if (math.isclose(0,c_n.x,abs_tol=0.01) and 
                    math.isclose(-1,c_n.y,rel_tol=0.1) and 
                    math.isclose(0,c_n.z,abs_tol=0.01)):
                    self.l_contact_norm=True
                    break
                else:
                    self.l_contact_norm=False
                    continue
            self.cancel_goal()
        return
    def cancel_goal(self):
        if self.r_contact_col_name==self.l_contact_col_name: #and self.r_contact_norm and self.l_contact_norm :
            self.process=False
            
            client.cancel_goal()
            #Robotiq.stop(client, block=False)
            
            #ATTACH LINK
            req = AttachRequest()
            req.model_name_1 = "robot"
            req.link_name_1 = "left_inner_finger"
            req.model_name_2 = self.l_contact_col_name.split("::")[0]
            req.link_name_2 = self.l_contact_col_name.split("::")[1]
            rospy.loginfo("Attaching {} and {}".format(req.model_name_2+"::"+req.link_name_2,"robot::left_inner_finger"))
            attach_srv.call(req)
        return
        	
def gripper_client(given):
    try:
        if given.data=="close":
            comparator.process=True
            rospy.loginfo('Closing Gripper')
            
            if sim:
                goal = control_msgs.msg.GripperCommandGoal()
                goal.command.position = 0.695  
                goal.command.max_effort = 5
                client.send_goal(goal)
            else:
                Robotiq.close(client, speed=0.05, force=5, block = False)
            
            if(client.wait_for_result(timeout=rospy.Duration(1))):
                return client.get_result()
            else:
                return ("stalled: True")
                
        if given.data=="open":
            comparator.process=False
            rospy.loginfo("Opening Gripper")
            if sim:
                #DETACH LINK
                try:
                    req = AttachRequest()
                    req.model_name_1 = "robot"
                    req.link_name_1 = "left_inner_finger"
                    req.model_name_2 = comparator.l_contact_col_name.split("::")[0]
                    req.link_name_2 = comparator.l_contact_col_name.split("::")[1]
                    rospy.loginfo("Detaching {} and {}".format(
                        req.model_name_2+"::"+req.link_name_2,
                        "robot::left_inner_finger"))
                    detach_srv.call(req)
                except:
                    rospy.loginfo("Link Not Detached")
                    
                comparator.r_contact_col_name=None
                comparator.r_contact_norm=None
                comparator.l_contact_col_name=None
                comparator.l_contact_norm=None
                
                
                goal = control_msgs.msg.GripperCommandGoal()
                goal.command.position = 0  
                goal.command.max_effort = 5
                client.send_goal(goal)
            else:
                Robotiq.open(client, speed=0.1, force=5, block = False)
                
            if(client.wait_for_result(timeout=rospy.Duration(1))):
                return client.get_result()
            else:
                return ("stalled: True")
    except rospy.ROSInterruptException:
        print ("Program interrupted before completion")
        
comparator= Comparator()
if sim:
    right_sub = rospy.Subscriber('/right_inner_finger_pad_bumper/', ContactsState, comparator.r_call)
    left_sub = rospy.Subscriber('/left_inner_finger_pad_bumper/', ContactsState, comparator.l_call)

gripper_sub = rospy.Subscriber('/gripper_controller/command', String, gripper_client)

rospy.loginfo('gripper action server initialized')
rospy.spin()



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.gt.pool.boards
## -- Module  : multicartpole
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-06  0.0.0     DA       Creation
## -- 2021-06-06  1.0.0     DA       Released first version
## -- 2021-09-11  1.0.0     MRD      Change Header information to match our new library name
## -- 2021-12-03  1.0.1     DA       Refactoring
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.1 (2021-12-03)

This module provides game board classes based on the Multi-CartPole environment
of the reinforcement learning pool.
"""

from mlpro.rl.models import Reward
from mlpro.rl.pool.envs.multicartpole import MultiCartPole
from mlpro.gt.models import *




## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class MultiCartPoleGT(MultiCartPole, GameBoard):

    """
    Game theoretical pendant for the reinforcement learning environment class MultiCartPole.
    """

    C_NAME          = 'MultiCartPole(GT)'

    def __init__(self, p_num_envs=2, p_logging=Log.C_LOG_ALL):
        MultiCartPole.__init__(self, p_num_envs=p_num_envs, p_reward_type=Reward.C_TYPE_EVERY_AGENT, p_logging=p_logging)






## -------------------------------------------------------------------------------------------------
## -------------------------------------------------------------------------------------------------
class MultiCartPolePGT(MultiCartPole, PGameBoard):
    """
    Potential Game theoretical pendant for the reinforcement learning environment class MultiCartPole.
    """

    C_NAME          = 'MultiCartPole(PGT)'

    def __init__(self, p_num_envs=2, p_logging=Log.C_LOG_ALL):
        MultiCartPole.__init__(self, p_num_envs=p_num_envs, p_reward_type=Reward.C_TYPE_EVERY_AGENT, p_logging=p_logging)



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_bf_002_timer.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-05-19  1.0.0     DA       Creation
## -- 2021-09-11  1.0.0     MRD      Change Header information to match our new library name
## -- 2021-11-13  1.0.1     DA       Minor fix
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.1 (2021-11-13)

This module demonstrates the Timer class functionality.
"""


import time
import random
from datetime import timedelta
from mlpro.bf.various import Timer, Log




# Demo class
class TimerDemo (Log):

    C_TYPE  = 'Demo class'
    C_NAME  = 'Timer'

    def __init__(self, p_timer:Timer):
        self.timer = p_timer
        super().__init__()


    def log(self, p_type, *p_args):
        super().log(p_type, 'Process time', self.timer.get_time(), ', Cycle', self.timer.get_lap_id(), 'Lap time', self.timer.get_lap_time(), '--', *p_args)


    def run_step(self, p_step_id):
        self.log(self.C_LOG_TYPE_I, 'Process step', p_step_id, 'started')
        duration = 0.6 * random.random()
        time.sleep(duration)
        self.log(self.C_LOG_TYPE_I, 'Process step', p_step_id, 'ended after', duration, 'seconds')


    def run_cycle(self):
        self.run_step(1) 
        self.run_step(2) 
        self.run_step(3) 
        if not self.timer.finish_lap():
            self.log(self.C_LOG_TYPE_W, 'Last process cycle timed out!!')


    def run(self):
        for i in range(10):
            self.run_cycle()




if __name__ == "__main__":

    # Example 1
    print('\n\n\nExample 1: Timer in virtual time mode with lap duration 1 day and 15 seconds...\n\n')
    t = Timer(Timer.C_MODE_VIRTUAL, timedelta(1,15,0))
    d = TimerDemo(t)
    d.run()


    # Example 2
    print('\n\n\nExample 2: Timer in real time mode with lap duration 1 second and forced timeout situations...\n\n')
    t = Timer(Timer.C_MODE_REAL, timedelta(0,1,0))
    d = TimerDemo(t)
    d.run()



## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_bf_005_hyperparameters.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-08-31  0.0.0     SY       Creation
## -- 2021-09-01  1.0.0     SY       Release of first version
## -- 2021-09-11  1.0.0     MRD      Change Header information to match our new library name
## -- 2021-12-10  1.0.1     DA       Refactoring, little beautifying
## -- 2022-02-25  1.0.2     SY       Refactoring due to auto generated ID in class Dimension
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.0.2 (2022-02-25)

This module demonstrates how to set-up hyperparameters using available HyperParamTuple, 
HyperParamSpace, and HyperParam classes.
"""



from mlpro.bf.ml import *


# 1 Setup a class that requires a tuple of hyperparameters 
class MyHyperparameter:
    
    def __init__(self):
        # 1.1 Construct a hyperparameter space using HyperParamSpace() and an empty tuple 
        self._hyperparam_space  = HyperParamSpace()
        self._hyperparam_tuple  = None
        self._init_hyperparam()
        

    def _init_hyperparam(self):
        # 1.2 Declare hyperparameters with unique id, names, and data type   
        self._hyperparam_space.add_dim(HyperParam('num_states','Z'))
        self._hyperparam_space.add_dim(HyperParam('smoothing','R'))
        self._hyperparam_space.add_dim(HyperParam('lr_rate','R'))
        self._hyperparam_space.add_dim(HyperParam('buffer_size','Z'))
        self._hyperparam_space.add_dim(HyperParam('update_rate','Z'))
        self._hyperparam_space.add_dim(HyperParam('sampling_size','Z'))
        self._hyperparam_tuple = HyperParamTuple(self._hyperparam_space)
        
        # 1.3 Set the hyperparameter with a default value 
        ids_ = self._hyperparam_tuple.get_dim_ids()
        self._hyperparam_tuple.set_value(ids_[0], 100)
        self._hyperparam_tuple.set_value(ids_[1], 0.035)
        self._hyperparam_tuple.set_value(ids_[2], 0.0001)
        self._hyperparam_tuple.set_value(ids_[3], 100000)
        self._hyperparam_tuple.set_value(ids_[4], 100)
        self._hyperparam_tuple.set_value(ids_[5], 256)


    def get_hyperparam(self) -> HyperParamTuple:
        return self._hyperparam_tuple
    



if __name__ == "__main__":

    # 2 Get value from the hyperparameter tuple 
    myParameter         = MyHyperparameter()
    for idx in myParameter.get_hyperparam().get_dim_ids():
        print('Variable with ID %s = %.2f'%(idx, myParameter.get_hyperparam().get_value(idx)))
        


    # 3 Overwrite current value with new desired value
    ids_ = myParameter.get_hyperparam().get_dim_ids()
    myParameter.get_hyperparam().set_value(ids_[0], 50)
    print('\nA new value for variable ID ids_[0]')
    print('Variable with ID ids_[0] = %.2f'%(myParameter.get_hyperparam().get_value(ids_[0])))


## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro
## -- Module  : howto_bf_004_store_plot_and_save_variables.py
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-06-16  1.0.0     SY       Creation/Release
## -- 2021-06-21  1.1.0     SY       Adjustment to updated DataPlotting class
## -- 2021-07-01  1.2.0     SY       Adjustment due to extension in save and load data
## -- 2021-09-11  1.2.1     MRD      Change Header information to match our new library name
## -- 2021-10-25  1.2.2     SY       Adjustment due to improvement in DataPlotting
## -- 2021-10-26  1.2.3     SY       Rename module
## -------------------------------------------------------------------------------------------------

"""
Ver. 1.2.3 (2021-10-26)

This module demonstrates how to store, plot, save and load variables.
"""


from mlpro.bf.various import *
from mlpro.bf.data import *
from mlpro.bf.plot import *
import random


if __name__ == "__main__":
    num_eps         = 10
    num_cycles      = 10000
    data_names      = ["reward","states_1","states_2","model_loss"]
    data_printing   = {"reward":        [True,0,10],
                       "states_1":      [True,0,4],
                       "states_2":      [True,0,4],
                       "model_loss":    [True,0,-1]}

    ## 1. How to store data ##
    mem = DataStoring(data_names)
    for ep in range(num_eps):
        ep_id = ("ep. %s"%str(ep+1))
        mem.add_frame(ep_id)
        for i in range(num_cycles):
            mem.memorize("reward",ep_id,random.uniform(0+(ep*0.5),5+(ep*0.5)))
            mem.memorize("states_1",ep_id,random.uniform(2-(ep*0.2),4-(ep*0.2)))
            mem.memorize("states_2",ep_id,random.uniform(0+(ep*0.2),2+(ep*0.2)))
            mem.memorize("model_loss",ep_id,random.uniform(0.25-(ep*0.02),1-(ep*0.07)))

    ## 2. How to plot stored data ##        
    # 2.1. Plotting data per cycle
    # mem_plot    = DataPlotting(mem, p_type=DataPlotting.C_PLOT_TYPE_CY, p_window=100,
    #                             p_showing=True, p_printing=data_printing, p_figsize=(7,7),
    #                             p_color="darkblue")
    # 2.2. Plotting data with continuous cycle
    mem_plot    = DataPlotting(mem, p_type=DataPlotting.C_PLOT_TYPE_EP, p_window=1000,
                                p_showing=True, p_printing=data_printing, p_figsize=(7,7),
                                p_color="darkblue")
    # 2.3. Plotting data per epsiode according to its mean value
    # mem_plot    = DataPlotting(mem, p_type=DataPlotting.C_PLOT_TYPE_EP_M, p_window=1,
    #                             p_showing=True, p_printing=data_printing, p_figsize=(7,7),
    #                             p_color="darkblue")
    mem_plot.get_plots()

    ## 3. How to save plots and data in binary file (variables, classes, etc.) ##
    path_save   = input("Input path_save : ")
    #Do not include quote-unquote ("" or '' ) into target path name
    mem_plot.save_plots(path_save, "pdf")
    mem_plot.save(path_save, "plot_memory")
    mem.save(path_save, "data_memory")
    mem.save_data(path_save, "data_storage", "\t")

    ## 4. How to load data from binary file ##
    path_load   = path_save
    mem_load    = DataStoring.load(path_load, "data_memory")
    print("Comparison :")
    print("Original data                : %.5f"%mem.memory_dict["reward"]["ep. 1"][0])
    print("Loaded data from binary file : %.5f"%mem_load.memory_dict["reward"]["ep. 1"][0])

    ## 5. How to load data from csv file ##
    data_names = []
    mem_from_csv = DataStoring(data_names)
    mem_from_csv.load_data(path_load, "data_storage.csv", "\t")
    print("Comparison :")
    print("Original data             : %.5f"%mem.memory_dict["reward"]["ep. 1"][0])
    print("Loaded data from csv file : %.5f"%mem_from_csv.memory_dict["reward"]["ep. 1"][0])




## -------------------------------------------------------------------------------------------------
## -- Project : MLPro - A Synoptic Framework for Standardized Machine Learning Tasks
## -- Package : mlpro.rl.envmodels
## -- Module  : mlp_robotinhtm
## -------------------------------------------------------------------------------------------------
## -- History :
## -- yyyy-mm-dd  Ver.      Auth.    Description
## -- 2021-12-17  0.0.0     MRD      Creation
## -- 2021-12-17  1.0.0     MRD      Released first version
## -- 2021-12-20  1.0.1     DA       Replaced 'done' by 'success'
## -- 2021-12-21  1.0.2     DA       Class MLPEnvMdel: renamed method reset() to _reset()
## -- 2022-01-02  2.0.0     MRD      Refactoring due to the changes on afct pool on
## --                                TorchAFctTrans
## -- 2022-02-25  2.0.1     SY       Refactoring due to auto generated ID in class Dimension
## -- 2022-05-22  2.0.2     MRD      Refactoring TorchAFct
## -- 2022-05-30  1.0.1     MRD      Cleaning up MLPEnvModel, now inherit directly from the
## --                                actual environment
## -------------------------------------------------------------------------------------------------

"""
Ver. 2.0.3 (2022-05-30)

This module provides Environment Model based on MLP Neural Network for
robotinhtm environment.
"""

import torch
import transformations

from mlpro.rl.models import *
from mlpro.rl.pool.envs.robotinhtm import RobotArm3D
from mlpro.rl.pool.envs.robotinhtm import RobotHTM
from mlpro.sl.pool.afct.afct_pytorch import TorchAFct

from torch.utils.data.sampler import SubsetRandomSampler
from collections import deque

def init(module, weight_init, bias_init, gain=1):
    weight_init(module.weight.data, gain=gain)
    bias_init(module.bias.data)
    return module

class RobotMLPModel(torch.nn.Module):
    def __init__(self, n_joint, timeStep):
        super(RobotMLPModel, self).__init__()
        self.n_joint = n_joint
        self.timeStep = timeStep
        self.hidden = 128

        init_ = lambda m: init(m, torch.nn.init.orthogonal_, lambda x: torch.nn.init.
                               constant_(x, 0), np.sqrt(2))

        self.model1 = torch.nn.Sequential(
            init_(torch.nn.Linear(self.n_joint,self.hidden)),
            torch.nn.Tanh(),
            init_(torch.nn.Linear(self.hidden,self.hidden)),
            torch.nn.Tanh(),
            init_(torch.nn.Linear(self.hidden,self.hidden)),
            torch.nn.Tanh(),
            init_(torch.nn.Linear(self.hidden,7*(self.n_joint+1))),
            torch.nn.Tanh()
            )
        
    def forward(self, I):
        BatchSize=I.shape[0]
        newI = I.reshape(BatchSize,2,self.n_joint) * torch.cat([torch.Tensor([self.timeStep]).repeat(1,self.n_joint), torch.ones(1,self.n_joint)])
        newI = torch.sum(newI,dim=1)
        out2 = self.model1(newI)   
        out2 = out2.reshape(BatchSize,self.n_joint+1,7)
        return out2

class IOElement(BufferElement):
    def __init__(self, p_input: torch.Tensor, p_output: torch.Tensor):

        super().__init__({"input": p_input, "output": p_output})


# Buffer
class MyOwnBuffer(Buffer, torch.utils.data.Dataset):
    def __init__(self, p_size=1):
        Buffer.__init__(self, p_size=p_size)
        self._internal_counter = 0

    def add_element(self, p_elem: BufferElement):
        Buffer.add_element(self, p_elem)
        self._internal_counter += 1

    def get_internal_counter(self):
        return self._internal_counter

    def __getitem__(self,idx):
        return self._data_buffer["input"][idx], self._data_buffer["output"][idx]

class RobothtmAFct(TorchAFct):
    C_NAME = "Robothtm Adaptive Function"
    C_BUFFER_CLS = MyOwnBuffer

    def _setup_model(self):
        self.joint_num = self._output_space.get_num_dim() - 6
        self.net_model = RobotMLPModel(self.joint_num, 0.01)
        self.optimizer = torch.optim.Adam(self.net_model.parameters(), lr=3e-4)
        self.loss_dyn = torch.nn.MSELoss()
        self.train_model = True
        self.input_temp = None

        self.sim_env = RobotArm3D()

        joints = []

        jointType = []
        vectLinkLength = [[0, 0, 0], [0, 0, 0]]
        jointType.append("rz")
        for joint in range(self.joint_num - 1):
            vectLinkLength.append([0, 0.7, 0])
            jointType.append("rx")

        jointType.append("f")

        for x in range(len(jointType)):
            vectorLink = dict(x=vectLinkLength[x][0], y=vectLinkLength[x][1], z=vectLinkLength[x][2])
            joint = dict(
                Joint_name="Joint %d" % x,
                Joint_type=jointType[x],
                Vector_link_length=vectorLink,
            )
            joints.append(joint)

        for robo in joints:
            self.sim_env.add_link_joint(
                lvector=torch.Tensor(
                    [
                        [
                            robo["Vector_link_length"]["x"],
                            robo["Vector_link_length"]["y"],
                            robo["Vector_link_length"]["z"],
                        ]
                    ]
                ),
                jointAxis=robo["Joint_type"],
                thetaInit=torch.Tensor([np.radians(0)]),
            )

        self.sim_env.update_joint_coords()

    def _input_preproc(self, p_input: torch.Tensor) -> torch.Tensor:
        input = torch.cat([p_input[0][6+self.joint_num:], p_input[0][6:6+self.joint_num]])
        input = input.reshape(1,self.joint_num*2)
        self.input_temp = p_input[0][:3].reshape(1,3)
        
        return input

    def _output_postproc(self, p_output: torch.Tensor) -> torch.Tensor:
        angles = torch.Tensor([])
        thets = torch.zeros(3)
        for idx in range(self.joint_num):
            angle = torch.Tensor(transformations.euler_from_quaternion(p_output[-1][idx][3:].detach().numpy(), axes="rxyz")) - thets
            thets = torch.Tensor(transformations.euler_from_quaternion(p_output[-1][idx][3:].detach().numpy(), axes="rxyz"))
            angles = torch.cat([angles, torch.norm(angle).reshape(1, 1)], dim=1)

        output = torch.cat([self.input_temp, p_output[-1][-1][:3].reshape(1,3)], dim=1)
        output = torch.cat([output, angles], dim=1)

        return output
    
    def _adapt(self, p_input: Element, p_output: Element) -> bool:
        model_input = deque(p_input.get_values()[6:])
        model_input.rotate(self.joint_num)
        model_input = torch.Tensor([list(model_input)])

        self.sim_env.set_theta(torch.Tensor([p_output.get_values()[6 : 6 + self.joint_num]]))
        self.sim_env.update_joint_coords()

        model_output = self.sim_env.convert_to_quaternion().reshape(1,self.joint_num+1,7)

        self._add_buffer(IOElement(model_input, model_output))

        if self._buffer.get_internal_counter() % 100 != 0:
            return False

        # Divide Test and Train
        if self.train_model:
            dataset_size = len(self._buffer)
            indices = list(range(dataset_size))
            split = int(np.floor(0.3 * dataset_size))
            np.random.seed(random.randint(1,1000))
            np.random.shuffle(indices)
            train_indices, test_indices = indices[split:], indices[:split]

            train_sampler = SubsetRandomSampler(train_indices)
            test_sampler = SubsetRandomSampler(test_indices)
            trainer = torch.utils.data.DataLoader(self._buffer, batch_size=100, sampler=train_sampler)
            tester = torch.utils.data.DataLoader(self._buffer, batch_size=100, sampler=test_sampler)

            # Training
            self.net_model.train()

            for i, (In, Label) in enumerate(trainer):
                outputs = self.net_model(In)
                loss = self.loss_dyn(outputs, Label)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

            test_loss = 0

            self.net_model.eval()
            for i, (In, Label) in enumerate(tester):
                outputs = self.net_model(In)
                loss = self.loss_dyn(outputs, Label)
                test_loss += loss.item()

            if test_loss/len(tester) < 5e-9:
                self.train_model = False

        return True

    def _add_buffer(self, p_buffer_element: IOElement):
        self._buffer.add_element(p_buffer_element)

class MLPEnvModel(RobotHTM, EnvModel):
    C_NAME = "MLP Env Model"

    def __init__(
        self,
        p_num_joints=4,
        p_target_mode="Random",
        p_ada=True,
        p_logging=False,
    ):

        RobotHTM.__init__(self, p_num_joints=p_num_joints, p_target_mode=p_target_mode)
        
        # Setup Adaptive Function
        # HTM Function Here
        afct_strans = AFctSTrans(
            RobothtmAFct,
            p_state_space=self._state_space,
            p_action_space=self._action_space,
            p_threshold=1.8,
            p_buffer_size=20000,
            p_ada=p_ada,
            p_logging=p_logging,
        )

        EnvModel.__init__(
            self,
            p_observation_space=self._state_space,
            p_action_space=self._action_space,
            p_latency=timedelta(seconds=self.dt),
            p_afct_strans=afct_strans,
            p_afct_reward=None,
            p_afct_success=None,
            p_afct_broken=None,
            p_ada=p_ada,
            p_logging=p_logging,
        )

        self.reset()

